import logging
import numpy as np
import torch
import matplotlib.pyplot as mplt
from torch.utils.data import TensorDataset
from sklearn.model_selection import train_test_split
from sklearn.cluster import KMeans
from fitxf.math.fit.utils.TensorUtils import TensorUtils
from fitxf.math.utils.Profile import Profiling


class FitUtils:

    @staticmethod
    def map_labels_to_consecutive_numbers(
            lbl_list,
    ):
        map_lbl_to_idx = FitUtils().map_labels_to_numbers(
            lbl_list = lbl_list,
            existing_label_map = None,
        )
        map_idx_to_lbl = {i:lbl for lbl,i in map_lbl_to_idx.items()}
        lbl_list_mapped = [map_lbl_to_idx[lbl] for lbl in lbl_list]
        return map_idx_to_lbl, map_lbl_to_idx, lbl_list_mapped

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        self.tensor_utils = TensorUtils(logger=self.logger)
        return

    def map_labels_to_numbers(
            self,
            lbl_list,
            existing_label_map = None,
            strict_1_to_1 = False,
    ):
        existing_label_map = {} if existing_label_map is None else existing_label_map

        existing_lbls = set(existing_label_map.keys())
        existing_idxs = set(existing_label_map.values())
        not_1_to_1 = len(existing_lbls) != len(existing_idxs)
        if not_1_to_1:
            errmsg = 'Unique labels (' + str(len(existing_lbls)) + ') != unique indexes (' + str(len(existing_idxs)) \
                     + ') in existing Label map ' + str(existing_label_map)
            self.logger.error(errmsg)
            if strict_1_to_1:
                raise Exception(errmsg)

        n_unique = len(np.unique(lbl_list + list(existing_lbls)))

        new_unique_lbls = list( set(lbl_list).difference(existing_lbls) )
        new_unique_lbls.sort()
        fill_indexes = list( set(range(n_unique)).difference(existing_idxs) )
        fill_indexes.sort()
        fill_indexes = fill_indexes[0:len(new_unique_lbls)]

        list_map_lbl_all = list(existing_label_map.items()) + list(zip(new_unique_lbls, fill_indexes))
        list_map_lbl_all.sort()
        self.logger.debug(
            'Existing labels: ' + str(existing_lbls) + ', new labels: ' + str(new_unique_lbls)
            + ', existing indexes: ' + str(existing_idxs) + ', new indexes: ' + str(fill_indexes)
            + ', list map lbl ' + str(list_map_lbl_all)
        )
        return {k: v for k, v in list_map_lbl_all}

    def get_point_distances(
            self,
            np_tensors: np.ndarray,
            # if not None, we take distances of points relative to this
            np_center = None,
            # permitted values 'exact', 'mc'
            method = 'exact',
            # permitted values 'euclid', 'euclid-square', 'cosine'
            metric = 'euclid',
            mc_iters = 10000,
    ):
        assert metric in ['euclid', 'euclid-square', 'cosine'], 'Metric not allowed "' + str(metric) + '"'
        # do a Monte-carlo
        distances = []
        n_points = len(np_tensors)
        if np_center is None:
            assert n_points >= 2, 'At least 2 points required for distances relative to each other'
        else:
            assert n_points >= 1, 'At least 1 point required for distances relative to center'

        self.logger.debug('Find by method "' + str(method) + '" distances between ' + str(n_points) + ' points')

        # Select pair indexes first, without calculating distance yet
        index_pairs = []
        if method == 'mc':
            while True:
                if len(index_pairs) >= mc_iters:
                    break
                a, b = np.random.randint(low=0, high=n_points, size=2)
                if a != b:
                    index_pairs.append((a, b))
        else:
            if np_center is None:
                for i in range(n_points):
                    [index_pairs.append((i, k)) for k in range(i + 1, n_points, 1)]
                assert len(index_pairs) == n_points * (n_points - 1) / 2
            else:
                [index_pairs.append((i, i)) for i in range(len(np_tensors))]

        self.logger.debug('Index pairs: ' + str(index_pairs))

        for a, b in index_pairs:
            if metric in ['euclid', 'euclid-square']:
                np_ref = np_tensors[b] if np_center is None else np_center
                euclid_square = np.sum( (np_tensors[a] - np_ref)**2 )
                distances.append(euclid_square) if metric == 'euclid-square' else distances.append(euclid_square ** 0.5)
            elif metric in ['cosine']:
                np_ref = np_tensors[b:(b+1)] if np_center is None else np_center.reshape([1]+list(np_center.shape))
                top_indexes, cos_dist = self.tensor_utils.dot_sim(
                    x = np_tensors[a:(a+1)],
                    ref = np_ref,
                    return_tensors = 'np',
                )
                self.logger.debug(
                    'Result cosine distance between index ' + str(a) + ' and index ' + str(b) + ', got ' + str(cos_dist)
                )
                distances.append(cos_dist)
            else:
                raise Exception('Unsupported metric "' + str(metric) + '"')
        return np.array(distances)

    def torch_train(
            self,
            model,
            loss_func,
            optimizer,
            # just an array of batch tuples, not necessarily the stupid overcomplicated torch DataLoader
            #    e.g. [(X_batch_1, y_batch_1), (X_batch_2, y_batch_2), ...]
            # or if attention masks included
            #    e.g. [(X_batch_1, attn_mask_batch_1, y_batch_1), (X_batch_2, attn_mask_batch_2, y_batch_2), ...]
            train_dataloader,
            epochs,
            regularization_type = None,
            regularization_lambda = 0.01,
            accum_steps = 1,
            plot_losses = False,
    ):
        profiler = Profiling(logger=self.logger)
        profiler_id = 'model=' + str(model) + ', epochs=' + str(epochs) + ', accum_steps=' + str(accum_steps)
        profiler.start_time_profiling(id=profiler_id)
        # Train mode means dropout layers are activated, gradients, etc
        model.train()
        losses = []
        # device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        print_epochs = np.round( np.linspace(0, epochs-1, 5), 0 )
        for epoch in range(epochs):
            run_loss = self.torch_train_1_epoch(
                epoch = epoch,
                dataloader = train_dataloader,
                model = model,
                loss_func = loss_func,
                regularization_type = regularization_type,
                regularization_lambda = regularization_lambda,
                optimizer = optimizer,
                accum_steps = accum_steps,
            )
            losses.append(run_loss)
            if epoch in print_epochs:
                pct = round(100 * epoch / (epochs-1), 0)
                self.logger.info(
                    str(pct) + '% done, losses over ' + str(epoch) + ' epochs, from ' + str(losses[0])
                    + ' to ' + str(losses[-1])
                )

        self.logger.debug('Losses over ' + str(epochs) + ' epochs: ' + str(losses))
        profiler.record_time_profiling(id=profiler_id, msg=profiler_id, logmsg=True)
        if plot_losses:
            mplt.plot(losses, linestyle='dotted')
            mplt.show()
        return losses

    def torch_train_1_epoch(
            self,
            epoch,
            # torch DataLoader
            dataloader,
            model,
            loss_func,
            optimizer,
            regularization_type = None,
            regularization_lambda = 0.01,
            # gradient accumulation steps
            # see https://wandb.ai/wandb_fc/tips/reports/How-To-Implement-Gradient-Accumulation-in-PyTorch--VmlldzoyMjMwOTk5
            accum_steps = 1,
    ):
        running_loss = 0.
        # Total step/batch should be int(n_train_samples/batch_size)
        for step, batch in enumerate(dataloader, 0):
            # batch = tuple(t.to(device) for t in batch)
            b_input_ids, b_input_mask, b_labels = batch

            start_accum = (step % accum_steps == 0)
            done_accum = ((step + 1) % accum_steps == 0) or (step + 1 == len(dataloader))

            if start_accum:
                # reset model paramater gradients before forward pass
                optimizer.zero_grad()

            # Forward pass
            pred_batch_y = model(b_input_ids)
            loss = loss_func(pred_batch_y, b_labels)
            loss = loss / accum_steps

            # Regularization
            # Apply L1 regularization
            if regularization_type == 'L1':
                l1_norm = sum(p.abs().sum() for p in model.parameters())
                reg_penalty = regularization_lambda * l1_norm
            # Apply L2 regularization
            elif regularization_type == 'L2':
                l2_norm = sum(p.pow(2).sum() for p in model.parameters())
                reg_penalty = regularization_lambda * l2_norm
            else:
                reg_penalty = 0
            self.logger.debug(
                '"' + str(regularization_type) + '" regularization penalty ' + str(reg_penalty) + ', loss from '
                + str(loss) + ' to new loss ' + str(loss+reg_penalty) + ', model parameters: '
                + str([p for p in model.parameters()])
            )
            loss += reg_penalty

            # Backward pass
            loss.backward()

            if done_accum:
                # adjust model parameter weights
                optimizer.step()

            running_loss += loss.item() * accum_steps
            self.logger.debug(
                'Epoch ' + str(epoch) + ': step ' + str(step) + ' loss ' + str(running_loss / 2000)
            )
        return running_loss

    def training_data_split(
            self,
            val_ratio,
            inputs,                 # torch.Tensor
            labels,                 # torch.Tensor
            attn_masks     = None,  # torch.Tensor
            # if False, means won't care if some labels only have 1 data
            stratify_label = False,
            return_tensors = 'pt',
            return_indexes = False,
    ):
        if val_ratio <= 0:
            # Return whole data set for both training/validation
            train_idx = np.arange(len(labels))
            val_idx = np.arange(len(labels))
        else:
            # Indices of the train and validation splits stratified by labels
            train_idx, val_idx = train_test_split(
                # We just need to return indexes of split, so just pass an array of 0 index to size of labels
                np.arange(len(labels)),
                test_size = val_ratio,
                shuffle   = True,
                # if None, means won't care if some labels only have 1 data
                stratify  = None if not stratify_label else labels,
                # stratify  = labels
            )
        # Train and validation sets
        if attn_masks is None:
            fake_mask = torch.ones(len(inputs[train_idx]))
            train_set = TensorDataset(inputs[train_idx], fake_mask, labels[train_idx])
            val_set = TensorDataset(inputs[val_idx], fake_mask, labels[val_idx])
        else:
            train_set = TensorDataset(inputs[train_idx], attn_masks[train_idx], labels[train_idx])
            val_set = TensorDataset(inputs[val_idx], attn_masks[val_idx], labels[val_idx])
        if return_indexes:
            return train_set, val_set, train_idx, val_idx
        else:
            return train_set, val_set

    def normalize_tensor_2D(
            self,
            # At least of size (n, m)
            X,
    ):
        x_size = X.size()
        assert len(x_size) == 2, '2D tensor expected but got ' + str(len(x_size)) + 'D'
        X_nm = torch.clone(X)
        tmp = torch.pow(X_nm, exponent=2)
        sum_sqrt = torch.sqrt(torch.sum(tmp, dim=-1))
        # print('***** Sum sqrt')
        # print(sum_sqrt)
        for i in range(X_nm.size()[0]):
            X_nm[i] = X_nm[i] / max(sum_sqrt[i], 1e-9)
            # assert torch.absolute(torch.sum(X_nm[i] * X_nm[i]) - 1.0) < 0.000001
        return X_nm

    def fit_mean(
            self,
            X,          # torch.Tensor
            # 0 index, unique valus in sequence 0, 1, 2, 3, ...
            labels,     # torch.Tensor
            # if >1, we use k-cluster to get the desired number of (cluster) "means"
            mean_per_label = 1,
            normalize_X = False,
            # Clone points if not enough
            clone_points = True,
    ):
        # Don't modify input Tensor X, thus we make a copy
        X_nm = torch.clone(X)
        if normalize_X:
            X_nm = self.normalize_tensor_2D(X=X)

        n_unique = len(torch.unique(labels))
        labels_unique = list(np.unique(labels))
        labels_unique.sort()
        list_0_to_n = list(range(n_unique))
        list_0_to_n.sort()
        assert labels_unique == list_0_to_n, 'Labels must be consecutive 0 to n, unique labels ' + str(labels_unique)
        self.logger.info('Fit mean with labels ' + str(labels) + ',X ' + str(X))

        mean_tensors = []
        for i_lbl in range(n_unique):
            # Indexes True/False of the desired label
            idx_tmp = labels == i_lbl
            x_tmp = X_nm[idx_tmp]
            assert x_tmp.size()[0] >= 1, 'Label ' + str(i_lbl) + ' must have at least 1 point'
            if clone_points:
                len_ori = x_tmp.size()[0]
                while x_tmp.size()[0] < mean_per_label:
                    x_tmp = torch.cat((x_tmp, x_tmp), dim=0)
                if x_tmp.size()[0] > len_ori:
                    self.logger.warning(
                        'Not enough points for label ' + str(i_lbl) + ', from ' + str(len_ori)
                        + ' points concatenated to ' + str(x_tmp.size()[0]) + ' points: ' + str(x_tmp)
                    )
            assert x_tmp.size()[0] >= mean_per_label, \
                str(x_tmp) + ' must have at least ' + str(mean_per_label) + ' points'
            # Mean of all tensors in this label, to create a single/few "representations" of the label
            if mean_per_label > 1:
                # Повторные точки могут вызывать предупреждение
                # "ConvergenceWarning: Number of distinct clusters found smaller than n_clusters."
                # Но мы это игнорируем
                # n_init должен быть достаточно большим чтобы результат сходился
                kmeans = KMeans(n_clusters=mean_per_label, n_init=10, random_state=0).fit(X=x_tmp)
                cluster_centers = kmeans.cluster_centers_
                # print(cluster_centers)
                mean_tensors.append(torch.from_numpy(cluster_centers))
                self.logger.info(
                    'Loop label ' + str(i_lbl) + ', mean per label ' + str(mean_per_label) + ', x ' + str(x_tmp)
                    + ', centers ' + str(cluster_centers)
                )
            else:
                mean_tmp = x_tmp.mean(dim=0)
                mean_tensors.append(mean_tmp)
        # Create a torch tensor from a list or torch tensors
        mean_tensors = torch.stack(mean_tensors)
        # make sure is float32
        return mean_tensors.type(torch.FloatTensor)

    def predict_mean(
            self,
            X,  # torch.Tensor
            mean_tensors,   # torch.Tensor
            reps_per_label,
            top_n     = 5,
            remove_dup_lbl = True,
            normalize_X    = False,
    ):
        # Don't modify input Tensor X, thus we make a copy
        X_nm = torch.clone(X)
        if normalize_X:
            X_nm = self.normalize_tensor_2D(X=X)
            metric = 'dot'
        else:
            metric = 'dist'

        # Resize if there are multiple representations per label
        size_tmp = list(mean_tensors.size())
        assert len(size_tmp) in [2,3], 'Allowed dimensions [2,3] for ' + str(size_tmp)
        if len(size_tmp) == 3:
            mean_tensors_flattenned = mean_tensors.reshape(
                size_tmp[0]*size_tmp[1],
                size_tmp[2]
            )
            self.logger.info(
                'Flattened mean tensors type ' + str(mean_tensors_flattenned.dtype)
                + ' from size ' + str(mean_tensors.size()) + ' to new size ' + str(mean_tensors_flattenned.size())
            )
        else:
            mean_tensors_flattenned = mean_tensors

        count_labels = mean_tensors.size()[0]

        if metric == 'dist':
            d = torch.FloatTensor()
            # loop calculate distance of all points to each representative label, then concatenate the rows
            for i in range(mean_tensors_flattenned.size()[0]):
                s = torch.sum((X_nm - mean_tensors_flattenned[i]) ** 2, dim=-1) ** 0.5
                s = s.reshape(1, s.size()[0])
                d = torch.cat((d, s), dim=0)
            pred_y = torch.t(d)
        else:
            tmp = torch.t(mean_tensors_flattenned)
            # If there are self.reps_per_label=3, then the correct label is floor(index/3)
            pred_y = torch.matmul(X_nm, tmp)
            # print('multiply tensors size ' + str(last_hidden_s_cls.size()) + ' with size ' + str(tmp.size()))

        if metric == 'dist':
            # min distance to max distance
            indexes_sorted = torch.argsort(pred_y, dim=-1, descending=False)
        else:
            # max dot product to min dot product
            indexes_sorted = torch.argsort(pred_y, dim=-1, descending=True)
        # If there are self.reps_per_label=3, then the correct label is floor(index/3)
        mapped_indexes_to_actual_labels = torch.floor(indexes_sorted / reps_per_label)

        # at this point all labels should appear the same number of times, exactly <reps_per_label> times

        # Remove duplicate labels in result if more than 1 mean_per_label
        # TODO what a mess this code, there should be a cleaner way
        if remove_dup_lbl:
            actual_labels_unique = torch.LongTensor()
            indexes_sorted_unique = torch.LongTensor()
            self.logger.debug(mapped_indexes_to_actual_labels)
            for i in range(mapped_indexes_to_actual_labels.size()[0]):
                row_actual_labels = mapped_indexes_to_actual_labels[i]
                row_indexes_sorted = indexes_sorted[i]
                j_positions = torch.LongTensor([0]*count_labels)
                for j in range(count_labels):
                    pos_j_all = (row_actual_labels == j).argwhere()
                    # all labels should appear <reps_per_label> times
                    assert len(pos_j_all) == reps_per_label, \
                        'row ' + str(i) + ' label ' + str(j) + ', inconsistency error'
                    # get first earliest position of j, by right should do torch.min()
                    pos_j = pos_j_all.flatten()[0]
                    j_positions[j] = pos_j
                    self.logger.debug('for row ' + str(row_actual_labels) + ' position of ' + str(j) + ': ' + str(pos_j))
                # sort from earliest positions to latest positions, so order is retained
                j_positions, _ = torch.sort(j_positions, descending=False, dim=0)
                tmp_actual_labels = row_actual_labels[j_positions]
                tmp_actual_labels = tmp_actual_labels.reshape(1, tmp_actual_labels.size()[0])
                tmp_indexes_sorted = row_indexes_sorted[j_positions]
                tmp_indexes_sorted = tmp_indexes_sorted.reshape(1, tmp_indexes_sorted.size()[0])
                # concatenate to final results
                actual_labels_unique = torch.cat((actual_labels_unique, tmp_actual_labels), dim=0)
                indexes_sorted_unique = torch.cat((indexes_sorted_unique, tmp_indexes_sorted), dim=0)
                # keep only those positions
                self.logger.debug('row ' + str(i) + ' j positions: ' + str(j_positions))
                self.logger.debug('row ' + str(i) + ' actual labels: ' + str(row_actual_labels) + ' --> ' + str(tmp_actual_labels))
                self.logger.debug('row ' + str(i) + ' index sorted: ' + str(row_indexes_sorted) + ' --> ' + str(tmp_indexes_sorted))
                self.logger.debug('row ' + str(i) + ' actual labels: ' + str(actual_labels_unique))
                self.logger.debug('row ' + str(i) + ' indexes sorted: ' + str(indexes_sorted_unique))
            indexes_final = indexes_sorted_unique
            actual_labels_final = actual_labels_unique
        else:
            indexes_final = indexes_sorted
            actual_labels_final = mapped_indexes_to_actual_labels

        probs = torch.FloatTensor()
        classes = torch.LongTensor()
        for i in range(indexes_final.shape[0]):
            p_row = pred_y[i][indexes_final[i]][:top_n]
            p_row = p_row.reshape(1, p_row.size()[0])
            c_row = actual_labels_final[i][:top_n]
            c_row = c_row.reshape(1, c_row.size()[0])
            probs = torch.cat((probs, p_row), dim=0)
            classes = torch.cat((classes, c_row), dim=0)
        return classes.type(torch.LongTensor), probs.type(torch.FloatTensor)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    # TODO Make as unit test below
    inputs = torch.Tensor([
        [0,1], [1,2], [2,3], [4,5], [5,6], [6,7], [7,8], [8,9], [9,10], [10,11],
    ])
    at_msk = torch.Tensor([True]*10)
    labels = torch.Tensor([0,0,0,0,0,1,1,1,2,2])
    fu = FitUtils()
    tset, vset = fu.training_data_split(
        inputs     = inputs,
        attn_masks = at_msk,
        labels     = labels,
        val_ratio  = 0.,
        stratify_label = True,
    )
    print('Training Set:')
    print(tset.tensors)
    print('Validation Set:')
    print(vset.tensors)

    exit(0)



import logging
import threading
from fitxf.math.utils.Profile import Profiling


#
# A key/value cache (bad naming below makes it "object"/"result")
#
class KeyValueCache:

    KEY_OBJECT = 'object'   # key
    KEY_RESULT = 'result'   # value
    KEY_REPEAT_COUNT = 'repeat_count'

    def __init__(
            self,
            cache_name = None,
            cache_size = 1000,
            # How much to remove from unproven cache when full, default is 50%
            rm_prop_when_full = 0.5,
            enable_text_similarity_lookup = True,
            text_similarity_fixed_len = 100,
            logger = None,
    ):
        self.cache_name = cache_name
        self.cache_size = cache_size
        assert self.cache_size > 0
        self.rm_prop_when_full = rm_prop_when_full
        self.enable_text_similarity_lookup = enable_text_similarity_lookup
        self.text_similarity_fixed_len = text_similarity_fixed_len
        self.logger = logger if logger is not None else logging.getLogger()

        self.cache = {}
        self.cache_stats = {'name': self.cache_name, 'total': 0, 'hit_exact': 0, 'hit_similar': 0, 'hit_rate': 0.0}

        self.textdiff = None
        if self.enable_text_similarity_lookup:
            # We cannot put on top of file, as there will be circular import
            from fitxf.math.lang.measures.TextDiff import TextDiff
            self.textdiff = TextDiff(logger=self.logger)
        self.ref_texts_keys = []
        self.ref_texts_chardiff_model = []

        self.__mutex_cache = threading.Lock()
        self.profiler = Profiling(logger=self.logger)

        self.logger.info(
            'Initialized SimpleCache with cache size ' + str(self.cache_size)
            + ', remove proportion when cache full ' + str(self.rm_prop_when_full)
            + ', text similarity lookup enabled = ' + str(self.enable_text_similarity_lookup)
        )
        return

    def update_cache_stats(
            self,
            hit_exact = 0,
            hit_similar = 0,
    ):
        assert hit_exact + hit_similar <= 1, 'Hit exact ' + str(hit_exact) + ', hit similar ' + str(hit_similar)
        # Total always add 1
        self.cache_stats['total'] += 1
        self.cache_stats['hit_exact'] += hit_exact
        self.cache_stats['hit_similar'] += hit_similar

        if self.cache_stats['total'] > 0:
            self.cache_stats['hit_rate'] = \
                (self.cache_stats['hit_exact'] + self.cache_stats['hit_similar']) / self.cache_stats['total']
        # Log stats every now and then
        if self.cache_stats['total'] % 500 == 0:
            self.logger.info('Key Value Cache "' + str(self.cache_name) + '" stats now: ' + str(self.cache_stats))
        return

    def derive_key(
            self,
            object,
    ):
        return str(object)

    # Look for <object> and return RESULT, something like lookup "key" and return "value", just naming problem
    def get_from_cache_threadsafe(
            self,
            object,
            # by default is exact key search. if this value >0, means will do a text similarity search
            similarity_threshold = 0.0,
    ):
        key = self.derive_key(object=object)
        hit_exact, hit_similar = 0, 0

        try:
            self.__mutex_cache.acquire()

            # First we try to look for exact match in proven cache
            if key in self.cache.keys():
                self.cache[key][self.KEY_REPEAT_COUNT] += 1
                self.logger.info(
                    'Found exact match in cache "' + str(key) + '": ' + str(self.cache[key])
                    + ' Text repeat count now ' + str(self.cache[key][self.KEY_REPEAT_COUNT])
                    + ' "' + str(key) + '"'
                )
                hit_exact = 1
                return self.cache[key][self.KEY_RESULT]
            elif similarity_threshold > 0.0:
                if not self.ref_texts_keys:
                    return None
                top_keys, top_distances = self.textdiff.text_similarity(
                    candidate_text = key,
                    ref_text_list = self.ref_texts_keys,
                    candidate_text_model = None,
                    ref_text_model_list = self.ref_texts_chardiff_model,
                    ref_str_len = self.text_similarity_fixed_len,
                    top_k = 5,
                    model = 'chardiff',
                )
                if top_keys:
                    if top_distances[0] <= similarity_threshold:
                        key_similar = top_keys[0]
                        self.logger.info(
                            'Found via similarity search for "' + str(key) + '", a similar key "' + str(key_similar)
                            + '" distance ' + str(top_distances[0])
                        )
                        self.cache[key_similar][self.KEY_REPEAT_COUNT] += 1
                        hit_similar = 1
                        return self.cache[key_similar][self.KEY_RESULT]
                return None
            else:
                return None
        # except Exception as ex:
        #     self.logger.error('Unexpected error: ' + str(ex))
        #     raise Exception(ex)
        finally:
            self.update_cache_stats(hit_exact=hit_exact, hit_similar=hit_similar)
            self.__mutex_cache.release()

    def add_to_cache_threadsafe(
            self,
            object, # key
            result, # value
    ):
        key = self.derive_key(object=object)

        try:
            self.__mutex_cache.acquire()

            is_cache_updated = False

            if len(self.cache) >= self.cache_size:
                cache_tmp = self.cache
                self.cache = {}
                # remove all with no hits first
                for k, v in cache_tmp.items():
                    if v[self.KEY_REPEAT_COUNT] > 0:
                        self.logger.info('Keep key "' + str(k) + '": ' + str(v))
                        self.cache[k] = v
                    else:
                        self.logger.info('Discard key "' + str(k) + '": ' + str(v))
                self.logger.info(
                    'Successfully cleanup up cache keeping only those with hits from length ' + str(len(cache_tmp))
                    + ', remaining items ' + str(len(self.cache))
                )
                is_cache_updated = True

                # If still not hit the desired target
                count_desired = round(self.cache_size * (1 - self.rm_prop_when_full))
                if len(self.cache) > count_desired:
                    # remove first added ones
                    count_thr = len(self.cache) - count_desired
                    # Keep only latest added ones
                    self.cache = {k:v for i,(k,v) in enumerate(self.cache.items()) if i >= count_thr}
                    self.logger.warning(
                        'Further clear hit cache to new size ' + str(len(self.cache)) + ', count thr ' + str(count_thr)
                        + ', max cache size ' + str(self.cache_size)
                    )

            if is_cache_updated:
                self.__update_text_model()

            if key not in self.cache.keys():
                self.cache[key] = {
                    self.KEY_OBJECT: object,
                    self.KEY_RESULT: result,
                    self.KEY_REPEAT_COUNT: 0,
                }
                self.ref_texts_keys.append(key)
                self.ref_texts_chardiff_model.append(
                    self.textdiff.get_text_model_chardiff(
                        text = key,
                        ref_str_len = self.text_similarity_fixed_len,
                    )
                )
            return key
        except Exception as ex:
            self.logger.error(
                'Error when adding key "' + str(key) + '" to cache: ' + str(ex)
            )
            return None
        finally:
            self.__mutex_cache.release()

    def __update_text_model(
            self,
    ):
        assert self.__mutex_cache.locked()

        self.ref_texts_keys = list(self.cache.keys())
        self.ref_texts_chardiff_model = [
            self.textdiff.get_text_model_chardiff(
                text = key,
                ref_str_len = self.text_similarity_fixed_len,
            )
            for key in self.cache.keys()
        ]
        self.logger.info('Successfully updated chardiff model')
        return

    def search_similar_object(
            self,
            text,
    ):
        try:
            self.__mutex_cache.acquire()
            return self.textdiff.text_similarity(
                candidate_text = text,
                ref_text_list = self.ref_texts_keys,
                candidate_text_model = None,
                ref_text_model_list = self.ref_texts_chardiff_model,
                ref_str_len = self.text_similarity_fixed_len,
                top_k = 5,
                model = 'chardiff',
            )
        finally:
            self.__mutex_cache.release()


class KeyValueCacheUnitTest:
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self):
        p = Profiling()
        data = [
            ('yo bro',           None,            ['yo bro']),
            ('dobriy dyen 1',    None,            ['yo bro', 'dobriy dyen 1']),
            ('kamusta 1',        None,            ['yo bro', 'dobriy dyen 1', 'kamusta 1']),
            ('dobriy dyen 1',    'dobriy dyen 1', ['yo bro', 'dobriy dyen 1', 'kamusta 1']), # exact hit
            # similarity search will return "dobriy dyen 1" above
            ('dobriy dyen 2',    'dobriy dyen 1', ['yo bro', 'dobriy dyen 1', 'kamusta 1', 'dobriy dyen 2']), # similar hit
            # At this point, we already reached unproven cache max size=4, and we clear those with no hits,
            # thus only ['dobriy dyen 1', 'kamusta 1'] remaining. 'kamusta 2' added after
            ('kamusta 2',        'kamusta 1',     ['dobriy dyen 1', 'kamusta 1', 'kamusta 2']), # similar hit
            ('kamusta 1',        'kamusta 1',     ['dobriy dyen 1', 'kamusta 1', 'kamusta 2']), # exact hit
            ('sami ludi 1',      None,            ['dobriy dyen 1', 'kamusta 1', 'kamusta 2', 'sami ludi 1']),
            # At this point, we already reached unproven cache max size=4, and we clear those with no hits,
            # thus only ['kamusta 1', 'sami ludi 1'] remaining. 'kamusta 2' added after
            ('sami ludi 1',      'sami ludi 1',   ['kamusta 1', 'sami ludi 1']), # exact hit
            ('sami ludi 2',      'sami ludi 1',   ['kamusta 1', 'sami ludi 1', 'sami ludi 2']), # similar hit
            ('sami ludi 3',      'sami ludi 1',   ['kamusta 1', 'sami ludi 1', 'sami ludi 2', 'sami ludi 3']), # similar hit
            # at this point cache will clear again keeping only those hit & latest ones added
            ('sami ludi 1',      'sami ludi 1',   ['kamusta 1', 'sami ludi 1']), # exact hit
        ]
        cache = KeyValueCache(
            cache_size = 4,
            rm_prop_when_full = 0.5,
            logger = self.logger,
        )

        count_hit_exact = 0
        count_hit_similar = 0

        for i, tp in enumerate(data):
            p.start_time_profiling(id='ut')

            print('At line #' + str(i+1) + '. ' + str(data[i]) + ', cache state ' + str(cache.cache))

            txt, expected_res_from_cache, expected_proven_cache_keys = tp
            res = cache.get_from_cache_threadsafe(
                object = txt,
                similarity_threshold = 0.2,
            )
            cache.add_to_cache_threadsafe(
                object = txt,
                result = txt,
            )
            if res is not None:
                if res == txt:
                    print('Exact hit with "' + str(res) + '" == "' + str(txt) + '"')
                    count_hit_exact += 1
                else:
                    print('Similar hit with "' + str(res) + '" != "' + str(txt) + '"')
                    count_hit_similar += 1
            # print('added key:', added_key, ', text', txt, ', result', res, ', expected result', expected_result)
            # print('#' + str(i) + '. Proven cache:\n' + str(cache.cache_proven) + ', unproven cache:\n' + str(cache.cache_unproven))
            assert res == expected_res_from_cache, \
                '#' + str(i+1) + '. Get result "' + str(res) + '" not expected result "' \
                + str(expected_res_from_cache) + '" for text "' + str(txt) + '", test data ' + str(tp)
            assert list(cache.cache.keys()) == expected_proven_cache_keys, \
                '#' + str(i+1) + '. Cache keys ' + str(list(cache.cache.keys())) \
                + '" not expected ' + str(expected_proven_cache_keys) + '" for text "' + str(txt) \
                + '", test data ' + str(tp)

            print('(AFTER ADD) At line #' + str(i+1) + '. cache state ' + str(cache.cache))

            top_keys, top_dists = cache.search_similar_object(text=txt)
            print('Sim search "' + str(txt) + '": ' + str(list(zip(top_keys, top_dists))))

            p.record_time_profiling(id='ut', msg=data[i], logmsg=True)

        print(cache.cache_stats)
        assert cache.cache_stats['total'] == len(data), \
            'Data length ' + str(len(data)) + ' but cache total ' + str(cache.cache_stats['total'])
        assert cache.cache_stats['hit_exact'] == count_hit_exact == 4, \
            'Exact hits ' + str(cache.cache_stats['hit_exact'])
        assert cache.cache_stats['hit_similar'] == count_hit_similar == 4, \
            'Similar hits ' + str(cache.cache_stats['hit_similar'])

        print('ALL TESTS PASSED')


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    KeyValueCacheUnitTest().test()

    exit(0)



import logging
import os
import re
import traceback
import numpy as np
import pandas as pd
import threading
from fitxf.math.utils.LockF import LockFile
from fitxf.math.datasource.DatastoreInterface import DatastoreInterface, DbParams, DatastoreInterfaceUnitTest
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Env import Env
from fitxf.math.utils.PkgVersion import PkgVersion


class Csv(DatastoreInterface):

    def __init__(
            self,
            db_params: DbParams,
            ignore_warnings = False,
            logger = None,
    ):
        super().__init__(
            db_params = db_params,
            ignore_warnings = ignore_warnings,
            logger = logger,
        )

        pkg_utils = PkgVersion(logger=self.logger)
        verdict, (self.py_maj_ver, self.py_min_ver)= pkg_utils.check_python(version="3.12", return_version=True)

        self.filepath = self.db_params.db_root_folder + os.sep + self.db_params.db_table

        self.use_file_lock = os.environ.get("CSV_USE_FILE_LOCK", "1").lower() in ("true", "1", "yes",)
        self.__lock_file_path = \
            db_params.db_root_folder + '/' + re.sub(pattern="[/\\\\:]", repl="_", string=self.db_params.db_table)
        self.__lock_file = LockFile(
            lock_file_path = self.__lock_file_path,
            logger = self.logger,
        )
        self.logger.info(
            'Using lock "' + str(self.use_file_lock) + '", file "' + str(self.__lock_file_path)
            + '" for file "' + str(self.filepath) + '"'
        )

        # This mutex will only work by thread/worker, for multithread/worker use file locks or RabbitMq
        self.__mutex = threading.Lock()
        return

    def __acquire_lock(self):
        if self.use_file_lock:
            self.__lock_file.acquire_file_cache_lock()
        self.__mutex.acquire()
        return

    def __release_lock(self):
        if self.use_file_lock:
            self.__lock_file.release_file_cache_lock()
        self.__mutex.release()
        return

    def connect(
            self,
    ):
        if not os.path.exists(self.filepath):
            self.logger.info(
                'Csv filepath does not exist "' + str(self.filepath) + '"'
            )
        return

    def __get_full_path(
            self,
            tablename,
    ):
        tablename = self.filepath if tablename is None else tablename

        if tablename != self.filepath:
            fullpath = self.db_params.db_root_folder + os.sep + str(tablename)
        else:
            fullpath = tablename
        assert fullpath == self.filepath
        return fullpath

    def __convert_match_phrase_to_tuples(
            self,
            match_phrase,
    ):
        if type(match_phrase) is dict:
            match_phrase_tuples = [(k, v) for k, v in match_phrase.items()]
        elif type(match_phrase) in [list, tuple]:
            for pair in match_phrase:
                assert len(pair) == 2, 'Match phrase tuple must be length 2: ' + str(pair)
            match_phrase_tuples = match_phrase
        else:
            raise Exception('Wrong type for match phrase ' + str(type(match_phrase)))

        return match_phrase_tuples

    def __get_dataframe_condition(
            self,
            match_phrase_tuples: list,
            match_condition: dict,
            df: pd.DataFrame,
    ):
        prm_cond_and = match_condition['and']
        prm_exact_match = match_condition['exact']

        condition = None
        for c, v in match_phrase_tuples:
            if prm_exact_match:
                condition_part = (df[c] == v)
            else:
                condition_part = df[c].str.contains(str(v), flags=re.IGNORECASE, regex=True, na=False)

            if prm_cond_and:
                condition = condition_part if condition is None else (condition & condition_part)
            else:
                condition = condition_part if condition is None else (condition | condition_part)
        return condition

    def get(
            self,
            # e.g. {"answer": "take_seat"}
            match_phrase,
            match_condition: dict = {'and': True, 'exact': True},
            tablename = None,
            request_timeout = 20.0,
    ):
        tablename = self.__get_full_path(tablename=tablename)

        match_phrase_tuples = self.__convert_match_phrase_to_tuples(match_phrase=match_phrase)

        try:
            self.__acquire_lock()
            df = pd.read_csv(tablename, index_col=False)
            self.logger.debug('Dataframe read from "' + str(tablename) + '", shape ' + str(df.shape))
            condition = self.__get_dataframe_condition(
                match_phrase_tuples = match_phrase_tuples,
                match_condition = match_condition,
                df = df,
            )
            df_filter = df[condition]
            return df_filter.to_dict(orient='records')
        except Exception as ex:
            self.logger.error(
                'Error reading data from table/index "' + str(tablename) + '": ' + str(ex)
                + ' Stack trace ' + str(traceback.format_exc())
            )
            return []
        finally:
            self.__release_lock()

    def get_all(
            self,
            key = None,
            max_records = 10000,
            tablename = None,
            request_timeout = 20.0,
    ):
        tablename = self.__get_full_path(tablename=tablename)

        if not os.path.exists(tablename):
            self.logger.error('Read path do not exist "' + str(tablename) + '"')
            return []
        try:
            self.__acquire_lock()
            df = pd.read_csv(tablename, index_col=False)
            self.logger.info('Dataframe read from "' + str(tablename) + '", shape ' + str(df.shape))
            return df.to_dict(orient='records')
        except Exception as ex:
            self.logger.error(
                'Error reading from csv filepath "' + str(tablename) + '": ' + str(ex)
                + ' Stack trace ' + str(traceback.format_exc())
            )
            return []
        finally:
            self.__release_lock()

    def get_indexes(self):
        return 'Delete files in folder yourself, this function not supported'

    def delete_index(
            self,
            tablename,
    ):
        tablename = self.__get_full_path(tablename=tablename)
        try:
            os.remove(tablename)
            self.logger.info('Deleted CSV index "' + str(tablename) + '"')
        except Exception as ex:
            self.logger.error(
                'Error deleting CSV index "' + str(tablename) + '": ' + str(ex)
                + ' Stack trace ' + str(traceback.format_exc())
            )
        return True

    def atomic_delete_add(
            self,
            delete_key: str,
            # list of dicts
            records: list[dict],
            tablename: str = None,
    ):
        tablename = self.__get_full_path(tablename=tablename)

        try:
            self.__acquire_lock()
            for rec in records:
                self.__delete(
                    match_phrase = rec,
                    match_condition = {'and': True, 'exact': True},
                    tablename = tablename,
                )
                return self.__add(
                    records = [rec],
                    tablename = tablename,
                )
        except Exception as ex:
            self.logger.error('Error occurred: ' + str(ex) + ' Stack trace ' + str(traceback.format_exc()))
            raise Exception(ex)
        finally:
            self.__release_lock()

    def add(
            self,
            # list of dicts
            records,
            tablename = None,
    ):
        tablename = self.__get_full_path(tablename=tablename)

        try:
            self.__acquire_lock()
            self.__add(
                records = records,
                tablename = tablename,
            )
        except Exception as ex:
            errmsg = \
                'Error occurred table "' + str(tablename) + '" add records ' + str(records) \
                + ', exception: ' + str(ex) + ' Stack trace ' + str(traceback.format_exc())
            self.logger.error(errmsg)
            raise Exception(errmsg)
        finally:
            self.__release_lock()

    def __add(
            self,
            # list of dicts
            records: list[dict],
            tablename: str,
    ):
        df_new = pd.DataFrame(records)
        new_record_columns = df_new.columns.tolist()

        if not os.path.exists(tablename):
            df_csv = pd.DataFrame(columns=new_record_columns)
            csv_columns = new_record_columns
            self.logger.info('Using new records columns: ' + str(new_record_columns))
        else:
            # this line will throw error if no columns or file is empty
            df_csv = pd.read_csv(tablename, index_col=False)
            existing_columns = df_csv.columns.tolist()
            csv_columns = existing_columns if len(existing_columns) > 0 else new_record_columns
            self.logger.info('Using columns: ' + str(csv_columns))

        missing_columns_in_records = [col for col in csv_columns if col not in new_record_columns]
        if missing_columns_in_records:
            self.logger.warning('Missing columns in records to be added: ' + str(missing_columns_in_records))
            for col in missing_columns_in_records:
                df_new[col] = None
                self.logger.warning('Added new column "' + str(col) + '" to new records/dataframe to be added')

        # make sure the columns are in the same order
        df_new = df_new[df_csv.columns]
        df = pd.concat([df_csv, df_new])
        df.to_csv(tablename, index=False)
        self.logger.info('Written total records ' + str(len(df)) + ' to file "' + str(tablename) + '"')
        return

    def delete(
            self,
            match_phrase,
            match_condition = {'and': True, 'exact': True},
            tablename = None,
    ):
        tablename = self.__get_full_path(tablename=tablename)
        try:
            self.__acquire_lock()
            self.__delete(
                match_phrase = match_phrase,
                match_condition = match_condition,
                tablename = tablename,
            )
        except Exception as ex:
            errmsg = \
                'Error occurred table "' + str(tablename) + '" delete ' + str(match_phrase) \
                + ', exception: ' + str(ex) + ' Stack trace ' + str(traceback.format_exc())
            self.logger.error(errmsg)
            raise Exception(errmsg)
        finally:
            self.__release_lock()

    def __delete(
            self,
            match_phrase: dict,
            tablename: str,
            match_condition: dict = {'and': True, 'exact': True},
    ):
        df = pd.read_csv(tablename, index_col=False)
        self.logger.debug('Dataframe read from "' + str(tablename) + '", shape ' + str(df.shape))

        match_phrase_tuples = self.__convert_match_phrase_to_tuples(match_phrase=match_phrase)

        condition = self.__get_dataframe_condition(
            match_phrase_tuples = match_phrase_tuples,
            match_condition = match_condition,
            df = df,
        )

        rows_removed = len(df[condition])
        df_remaining = df[np.logical_not(condition)]
        df_remaining.to_csv(tablename, index=False)
        return {'deleted': rows_removed}

    def delete_all(
            self,
            tablename: str = None,
    ):
        tablename = self.__get_full_path(tablename=tablename)
        try:
            self.__acquire_lock()
            self.__delete_all(
                tablename = tablename,
            )
        except Exception as ex:
            errmsg = \
                ('Error occurred table "' + str(tablename) + '" delete all, exception: ' + str(ex)
                 + ' Stack trace ' + str(traceback.format_exc()))
            self.logger.error(errmsg)
            raise Exception(errmsg)
        finally:
            self.__release_lock()

    def __delete_all(
            self,
            tablename: str,
    ):
        df = pd.read_csv(tablename, index_col=False)
        self.logger.debug('Dataframe read from "' + str(tablename) + '", shape ' + str(df.shape))

        condition = np.array([True]*len(df))

        rows_removed = len(df[condition])
        df_remaining = df[np.logical_not(condition)]
        df_remaining.to_csv(tablename, index=False)
        return {'deleted': rows_removed}

    def add_column(
            self,
            colnew,
            data_type = str,
            tablename = None,
            default_value = None,
    ):
        tablename = self.__get_full_path(tablename=tablename)

        colnew = str(colnew).strip()
        try:
            self.__acquire_lock()
            df = pd.read_csv(tablename, index_col=False)
            if colnew in df.columns:
                self.logger.info('Column already exists "' + str(colnew) + '"')
                return False
            df[colnew] = default_value
            df[colnew] = df[colnew].astype(dtype=data_type)
            df.to_csv(tablename, index=False)
            self.logger.info('Column successfully added "' + str(colnew) + '" as type "' + str(data_type) + '"')
        except Exception as ex:
            self.logger.error('Error occurred: ' + str(ex) + ' Stack trace ' + str(traceback.format_exc()))
        finally:
            self.__release_lock()

    # When csv writes a numpy array into file, it becomes a string, e.g.
    # '[ 2.32330292e-01 -6.20948195e-01  8.89505982e-01  6.19655311e-01\n -1.97012693e-01... ]'
    def convert_csv_string_array_to_float_array(
            self,
            string_array,
            custom_chars_remove = (),
    ):
        value = string_array.strip()
        # remove front/end brackets "[" and "]"
        if (self.py_maj_ver == 3) and (self.py_min_ver <= 11):
            value = re.sub(pattern="[\[\]]", repl="", string=value)
        else:
            value = re.sub(pattern="[[]]", repl="", string=value)
        # replace all tabs, newline, etc with space
        value = re.sub(pattern="[ \t\n\r]", repl=" ", string=value)
        # replace multiple spaces with a single space
        value = re.sub(pattern="[ ]+", repl=" ", string=value)
        # replace other custom characters
        for c in custom_chars_remove:
            value = re.sub(pattern="["+c+"]+", repl="", string=value)

        value = value.strip()
        value = value.split(sep=' ')
        float_array = [float(v) for v in value]
        return float_array


if __name__ == '__main__':
    Env.set_env_vars_from_file(env_filepath=Env().REPO_DIR + os.sep + '.env.fitxf.math.ut')
    DatastoreInterfaceUnitTest(
        ChildClass = Csv,
        logger = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    ).test(
        tablename = 'csvtest',
    )
    exit(0)



import torch
import logging
import os
from fitxf.math.lang.encode.LangModelInterface import LangModelInterface
from transformers import AutoTokenizer, AutoModel
from sentence_transformers import SentenceTransformer
from fitxf.math.utils.Env import Env
from fitxf.math.utils.Logging import Logging


#
# Leader Board: https://huggingface.co/spaces/mteb/leaderboard
#
class LangModelPt(LangModelInterface):

    DEFAULT_MODEL = 'sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2'

    def __init__(
            self,
            model_name: str = None,
            cache_folder: str = None,
            include_tokenizer: bool = False,
            logger = None,
    ):
        super().__init__(
            cache_folder = cache_folder,
            model_name = model_name,
            include_tokenizer = include_tokenizer,
            logger = logger,
        )

        self.model_name = self.DEFAULT_MODEL if self.model_name is None else self.model_name
        self.logger.info(
            'Model name "' + str(self.model_name) + '", cache folder "' + str(self.cache_folder) + '"'
        )

        # User may pass in model downloaded path
        if os.path.isdir(str(self.model_name)):
            self.model_path = self.model_name
        else:
            self.model_path = self.cache_folder + '/' + self.model_name

        assert os.path.isdir(self.model_path), 'Not a directory "' + str(self.model_path) + '"'
        self.logger.info('Model name "' + str(self.model_name) + '" path "' + str(self.model_path) + '"')

        self.logger.info(
            'Lang model "' + str(self.model_name) + '" with cache folder "' + str(self.cache_folder)
            + '", include tokenizer "' + str(self.include_tokenizer)
            + '", name_or_path "' + str(self.model_path) + '"'
        )
        if self.include_tokenizer:
            self.tokenizer = AutoTokenizer.from_pretrained(
                pretrained_model_name_or_path = self.model_path,
                # cache_folder = self.cache_folder,
            )
            self.logger.info(
                'OK Tokenizer for model "' + str(self.model_path) + '", cache folder "' + str(self.cache_folder) + '"'
            )
            self.model = AutoModel.from_pretrained(
                # hugging face will know to use the cache folder above, without specifying here it seems
                pretrained_model_name_or_path = self.model_path
            )
            self.logger.info(
                'OK Model "' + str(self.model_path) + '", cache folder "' + str(self.cache_folder) + '"'
            )
        else:
            self.model = SentenceTransformer(
                model_name_or_path = self.model_path,
                cache_folder = self.cache_folder,
            )
        return

    # Mean Pooling - Take attention mask into account for correct averaging
    def mean_pooling(self, model_output, attention_mask):
        token_embeddings = model_output[0]  # First element of model_output contains all token embeddings
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)

    def encode(
            self,
            content_list,
            # max length has got no meaning
            maxlen = None,
            return_tensors = 'pt',
            # does not apply here since we can't see the tokenization
            return_attnmasks = False,
            params_other = None,
    ):
        if self.include_tokenizer:
            self.logger.debug('Calculating embedding using tokenizer')
            embeddings_tensor = None
            for txt in content_list:
                # Tokenize sentences
                encoded_input = self.tokenizer(txt, padding=True, truncation=True, return_tensors='pt')

                # Compute token embeddings
                with torch.no_grad():
                    model_output = self.model(**encoded_input)

                # Perform pooling. In this case, max pooling.
                attn_masks = encoded_input['attention_mask']
                embedding = self.mean_pooling(model_output=model_output, attention_mask=attn_masks)
                if embeddings_tensor is not None:
                    embeddings_tensor = torch.cat((embeddings_tensor, embedding), dim=0)
                else:
                    embeddings_tensor = embedding

            attn_masks = None
            if return_tensors == 'pt':
                return (embeddings_tensor, attn_masks) if return_attnmasks else embeddings_tensor
            else:
                embedding_np = embeddings_tensor.cpu().detach().numpy()
                return (embedding_np, attn_masks) if return_attnmasks else embedding_np
        else:
            self.logger.debug('Calculating embedding using single sentence transformer wrapper')
            embedding = self.model.encode(
                sentences = content_list,
            )
            # TODO
            #    How to get this? Since this depends on the tokenizer used by the language model,
            #    this means we cannot calculate on our own by tokenizing.
            attn_masks = None

            if return_tensors == 'pt':
                pt_array = torch.from_numpy(embedding)
                return (pt_array, attn_masks) if return_attnmasks else pt_array
            else:
                return (embedding, attn_masks) if return_attnmasks else embedding


if __name__ == '__main__':
    er = Env()
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)

    lm = LangModelPt(
        model_name = 'intfloat/multilingual-e5-small',
        cache_folder = er.MODELS_PRETRAINED_DIR,
        logger = lgr,
    )
    text_list = [
        '칠리 페퍼', '와사비', '머스타드',
        '케이크', '도넛', '아이스크림',
    ]
    labels = ['hot', 'hot', 'hot', 'sweet', 'sweet', 'sweet']
    lm.visualize_embedding(
        encoding_np = lm.encode(content_list=text_list, return_tensors='np'),
        labels_list = labels,
    )

    print('rps', lm.speed_test(sentences=text_list, min_rounds=200))
    exit(0)



import logging
import os
from io import StringIO
import numpy as np
import pandas as pd
from fitxf.math.algo.encoding.Base64 import Base64
from fitxf.math.data.ut.LabelTextEmbed01 import DATA_LABEL_TEXT_EMBEDDING_01_TRAIN, DATA_LABEL_TEXT_EMBEDDING_01_EVAL
from fitxf.math.fit.transform.FitXformInterface import FitXformInterface
from fitxf.math.fit.transform.FitXformPca import FitXformPca
from fitxf.math.fit.transform.FitXformCluster import FitXformCluster, FitXformClusterCosine
from fitxf.math.utils.Env import Env
from fitxf.math.utils.Logging import Logging


class FitXformUnitTest:

    def __init__(
            self,
            lm_cache_folder = None,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        self.lm_cache_folder = lm_cache_folder
        self.base64 = Base64(logger=self.logger)
        return

    def test(self):
        for F, score_thr, ret_full_rec in [
            (FitXformPca, 0.95, False),
            (FitXformPca, 0.95, True),
            (FitXformCluster, 0.95, False),
            (FitXformClusterCosine, 0.95, False),
        ]:
            self.__test_fit(
                fitter_name = str(F.__class__),
                FitterClassType = F,
                avg_score_threshold = score_thr,
                ret_full_rec = ret_full_rec,
            )

        # Test pass thru mode
        x = np.random.rand(10,10)
        fitter = FitXformCluster(logger=self.logger)
        res = fitter.fit(
            X = x,
            test_mode = True,
        )
        n_centers = res[FitXformInterface.KEY_N_COMPONENTS_OR_CENTERS]
        centers = res[FitXformInterface.KEY_CENTERS]
        # self.logger.info('Test mode fit result for cluster: ' + str(res))
        assert n_centers == len(x)
        assert np.sum((x - centers)**2) < 0.0000000001

        pred_lbls, pred_probs = fitter.predict(X=x)
        for i, pred_lbls_row in enumerate(pred_lbls):
            top_pred_i, top_prob_i = pred_lbls_row[0], pred_probs[i][0]
            print(i, top_pred_i, top_prob_i)
            assert top_pred_i['cluster_label'] == top_pred_i['user_label_estimate'] == i
            assert top_prob_i**2 - 1 < 0.0000000001

        print('ALL TESTS PASSED OK')
        return

    def __test_fit(
            self,
            fitter_name: str,
            FitterClassType: type(FitXformInterface),
            avg_score_threshold: float,
            ret_full_rec: bool,
    ):
        def get_fitter_object() -> FitXformInterface:
            return FitterClassType(logger=self.logger)
        fitter = get_fitter_object()
        def get_data(
                s,
        ):
            df = pd.read_csv(
                filepath_or_buffer = StringIO(s),
                sep = ',',
                index_col = False,
            )
            columns_keep = ['label', 'text', 'embedding']
            for c in columns_keep: assert c in df.columns
            df = df[columns_keep]
            df.dropna(inplace=True)
            # _, _, df[self.col_label_std] = FitUtils().map_labels_to_consecutive_numbers(lbl_list=list(df[self.col_label]))
            self.logger.info('Successfully read data of shape ' + str(df.shape))
            return df

        df_train = get_data(s=DATA_LABEL_TEXT_EMBEDDING_01_TRAIN)
        df_eval = get_data(s=DATA_LABEL_TEXT_EMBEDDING_01_EVAL)

        texts_train, labels_train = df_train['text'].tolist(), df_train['label'].tolist()
        texts_eval, labels_eval = df_eval['text'].tolist(), df_eval['label'].tolist()
        full_recs_train = [
            {k: v for (k, v) in r.items() if k not in ['embedding']} for r in df_train.to_dict(orient='records')
        ]
        full_recs_eval = [
            {k: v for (k, v) in r.items() if k not in ['embedding']} for r in df_eval.to_dict(orient='records')
        ]

        try:
            raise Exception('Force to use pre-calculated embeddings.')
            from fitxf.math.lang.encode.LangModelPt import LangModelPt
            from fitxf.math.lang.encode.LangModelPtSingleton import LangModelPtSingleton
            def get_lm() -> LmPt:
                return LangModelPtSingleton.get_singleton(
                    LmClass = LangModelPt,
                    cache_folder = self.lm_cache_folder,
                    logger = self.logger,
                )
            # emb_train = get_lm().encode(text_list=texts_train, return_tensors='np')
            # emb_eval = get_lm().encode(text_list=texts_test, return_tensors='np')
        except Exception as ex:
            self.logger.info('Failed to calculate embeddings: ' + str(ex) + ', using precalculated embeddings instead.')
            emb_train = np.array([
                self.base64.decode_base64_string_to_numpy_array(s64=s, data_type=np.float64)
                for s in df_train['embedding'].tolist()
            ])
            emb_eval = np.array([
                self.base64.decode_base64_string_to_numpy_array(s64=s, data_type=np.float64)
                for s in df_eval['embedding'].tolist()
            ])
            # Remove some points, and add additional points for fine-tuning
            n_remove = 3
            emb_fine_tune = emb_train[:-n_remove]
            emb_fine_tune = np.append(emb_fine_tune, emb_eval, axis=0)
            labels_fine_tune = labels_train[:-n_remove]
            labels_fine_tune += labels_eval
            full_recs_fine_tune = full_recs_train[:-n_remove]
            full_recs_fine_tune += full_recs_eval

        # x = np.array([
        #     [1, 1, 1, 1], [2, 2, 2, 2],
        #     [2, 1.5, -1, 0.3], [1, 2, -2, 1],
        #     [3, 0.5, 0, -1], [1, 1, 1, -2],
        # ])
        res = fitter.fit_optimal(
            X = emb_train,
            X_labels = labels_train,
            X_full_records = full_recs_train,
            target_grid_density = 2.,
            measure = 'min',
            min_components = 3,
            # estimate n centers at most 2x number of unique labels
            max_components = len(np.unique(labels_train)) * 2,
        )

        self.__test_predictions(
            fitter = fitter,
            emb_train = emb_train,
            emb_eval = emb_eval,
            ret_full_rec = ret_full_rec,
            avg_score_threshold = avg_score_threshold,
            expected_top_labels = labels_eval,
        )
        # now load a new instance by saving json and reloading
        fitter_new_loaded = self.__test_save_load_model(
            fitter_old = fitter,
            FitterClassType = FitterClassType,
        )
        # do the same test on this new object, to make sure it works just like the original
        self.__test_predictions(
            fitter = fitter_new_loaded,
            emb_train = emb_train,
            emb_eval = emb_eval,
            # we no longer have full records
            ret_full_rec = False,
            avg_score_threshold = avg_score_threshold,
            expected_top_labels = labels_eval,
        )
        # now test fine tuning
        self.test_fine_tune(
            fitter = fitter,
            X = emb_fine_tune,
            X_labels = labels_fine_tune,
            X_full_records = full_recs_fine_tune,
            n_components = fitter.model_n_components_or_centers,
        )
        return

    def __test_predictions(
            self,
            fitter: FitXformInterface,
            emb_train: np.ndarray,
            emb_eval: np.ndarray,
            ret_full_rec: bool,
            avg_score_threshold: float,
            expected_top_labels: list,
    ):
        x_transform = fitter.X_transform
        x_transform_check = fitter.transform(X=emb_train)
        x_inverse_transform = fitter.inverse_transform(X=x_transform)

        # Check if estimation of actual value is correct
        diff = x_inverse_transform - emb_train
        sq_err_per_vect = np.sum(diff*diff) / len(diff)
        x_dim = emb_train.shape[-1]
        sq_err_per_vect_thr = 0.1 * x_dim
        assert sq_err_per_vect < sq_err_per_vect_thr, \
            '[' + str(fitter.__class__) + '] Estimate back using PCA, per vector sq err ' + str(sq_err_per_vect) \
            + '>=' + str(sq_err_per_vect_thr) + ', details: ' + str(diff)

        # Check if our manual calculation of the principal component values are correct
        diff = x_transform_check - x_transform
        sq_err_sum = np.sum(diff*diff)
        assert sq_err_sum < 0.000000001, \
            '[' + str(fitter.__class__) + '] Manual calculate PCA component values, sq err ' + str(sq_err_sum) \
            + ', diff ' + str(diff)

        for use_grid in (False, True,):
            pred_labels, pred_probs = fitter.predict(
                X = emb_eval,
                use_grid = use_grid,
                return_full_record = ret_full_rec,
                top_k = 2,
            )

            self.logger.info(
                '[' + str(fitter.__class__) + '] Use grid "' + str(use_grid)
                + '", predicted labels: ' + str(pred_labels)
            )

            scores = []
            for i, exp_lbl in enumerate(expected_top_labels):
                pred_top_label = pred_labels[i][0]['label'] if ret_full_rec else pred_labels[i][0]
                pred_top_label_2 = pred_labels[i][1]['label'] if ret_full_rec else pred_labels[i][1]
                if type(fitter) in [FitXformClusterCosine, FitXformCluster]:
                    # First in tuple is predicted cluster number, take
                    pred_top_cluster = pred_top_label['cluster_label']
                    pred_top_label = pred_top_label['user_label_estimate']
                    pred_top_cluster_2 = pred_top_label_2['cluster_label']
                    pred_top_label_2 = pred_top_label_2['user_label_estimate']

                    #
                    # Example of how to obtain data to zooom into search
                    #
                    condition_zoom = fitter.X_transform == pred_top_cluster
                    emb_zoom_local_space = emb_train[condition_zoom]
                    labels_zoom_local_space = [
                        lbl for i_lbl, lbl in enumerate(fitter.X_labels) if condition_zoom[i_lbl]
                    ]
                    assert len(emb_zoom_local_space) == len(labels_zoom_local_space)
                    self.logger.info('labels list zoom ' + str(labels_zoom_local_space))

                    zoom_lbls, zoom_probs = fitter.predict(
                        X = emb_eval[i:(i+1)],
                        X_search_local_space = emb_zoom_local_space,
                        labels_search_local_space = labels_zoom_local_space,
                        top_k = 3,
                    )
                    self.logger.info('Zoom search result ' + str(zoom_lbls) + ', probs ' + str(zoom_probs))
                    # raise Exception('asdf')

                # 1.0 for being 1st, 0.5 for appearing 2nd
                score_i = 1*(pred_top_label == exp_lbl) + 0.5*(pred_top_label_2 == exp_lbl)
                score_i = min(score_i, 1.0)
                scores.append(score_i)
                # Check only top prediction
                if pred_top_label != exp_lbl:
                    self.logger.warning(
                        '[' + str(fitter.__class__) + '] #' + str(i) + ' Use grid "' + str(use_grid)
                        + '". Predicted top label "' + str(pred_top_label) + '" not expected "' + str(exp_lbl) + '"'
                    )
            score_avg = np.mean(np.array(scores))
            self.logger.info(
                '[' + str(fitter.__class__) + '] Use grid "' + str(use_grid) + '". Mean score '
                + str(score_avg) + ', scores' + str(scores)
            )
            assert score_avg > avg_score_threshold, \
                '[' + str(fitter.__class__) + '] Use grid "' + str(use_grid) + '". Mean score fail ' + str(score_avg) \
                + ' < ' + str(avg_score_threshold) + '. Scores ' + str(scores)

    def __test_save_load_model(
            self,
            fitter_old: FitXformInterface,
            FitterClassType: type(FitXformInterface),
    ):
        #
        # Test model saving to json and reloading is ok
        #
        centers_before = np.array(fitter_old.model_centers)
        pca_before = np.array(fitter_old.model_principal_components)
        model_dict_b64jsondump = fitter_old.model_to_b64json(numpy_to_base64_str=True, dump_to_b64json_str=True)
        # model_dict = json.loads(model_dict_jsondump)
        # [self.logger.info(str(k) + ': ' + str(v)) for k, v in model_dict.items()]

        def get_fitter() -> FitXformInterface:
            return FitterClassType(logger=self.logger)

        fitter_new = get_fitter()
        fitter_new.load_model_from_b64json(model_b64json=model_dict_b64jsondump)

        assert fitter_new.model_centers.shape == centers_before.shape, \
            'Shape before ' + str(centers_before.shape) + ' but after ' + str(fitter_new.model_centers.shape)
        assert fitter_new.model_principal_components.shape == pca_before.shape, \
            'Shape before ' + str(pca_before.shape) + ' but after ' + str(fitter_new.model_principal_components.shape)

        diff = np.sum( (fitter_new.model_centers - centers_before) ** 2 )
        assert diff < 0.0000000001, \
                'Centers after and before different, before\n' + str(centers_before) \
                + ', after\n' + str(fitter_new.model_centers)
        diff = np.sum( (fitter_new.model_principal_components - pca_before) ** 2 )
        assert diff < 0.0000000001, \
                'Principle components after and before different, before\n' + str(pca_before) \
                + ', after\n' + str(fitter_new.model_principal_components)
        return fitter_new

    def test_fine_tune(
            self,
            fitter: FitXformInterface,
            X: np.ndarray,
            X_labels: list,
            X_full_records: list,
            n_components: int,
    ):
        res = fitter.fine_tune(
            X = X,
            X_labels = X_labels,
            X_full_records = X_full_records,
            n_components = n_components,
        )
        self.logger.info('Result of fine tune..')
        [self.logger.info(str(k) + ':' + str(v)) for k, v in res.items()]
        return


if __name__ == '__main__':
    FitXformUnitTest(
        lm_cache_folder = Env().MODELS_PRETRAINED_DIR,
        logger = Logging.get_default_logger(log_level=logging.INFO, propagate=False),
    ).test()
    exit(0)



import numpy as np
import logging
from fitxf.math.fit.utils.FitUtils import FitUtils
from fitxf.math.fit.utils.TensorUtils import TensorUtils
from fitxf.math.fit.cluster.Metrics import Metrics as ClusterMetrics
from sklearn.cluster import KMeans, MeanShift, DBSCAN
import matplotlib.pyplot as mplt


#
# The idea is this:
#   Case 1: All clusters (think towns) are almost-equally spaced apart
#      - in this case, suppose optimal cluster centers=n (think salesmen)
#      - if number of clusters k<n, then each salesman need to cover a larger area, and their average distances from each other is smaller
#      - if number of clusters k>n, then things become a bit crowded, with more than 1 salesman covering a single town
#      Thus at transition from n --> n+1 clusters, the average distance between cluster centers will decrease
#   Case 2: Some clusters are spaced much larger apart from other clusters
#      In this case, there will be multiple turning points, and we may take an earlier turning point or later turning points
#
# Нет кластеров - как узнать
#    1.
#
class Cluster:

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        self.fit_utils = FitUtils(logger=self.logger)
        self.tensor_utils = TensorUtils(logger=self.logger)
        return

    def estimate_min_max_clusters(
            self,
            n,
    ):
        max_clusters = 3*int(np.log(n))
        min_clusters = max(2, int(np.log(n)))
        self.logger.info(
            'Min/max clusters estimated as ' + str(min_clusters) + ' and ' + str(max_clusters) + ', n=' + str(n)
        )
        return min_clusters, max_clusters

    def derive_additional_cluster_info(
            self,
            x: np.ndarray,
            n_centers: int,
            cluster_centers: np.ndarray,
            cluster_labels: np.ndarray,
            metric,
    ):
        assert len(cluster_centers) > 0
        assert type(cluster_centers) is np.ndarray
        # get distances between all center pairs
        distances_cluster_centers = self.fit_utils.get_point_distances(
            np_tensors = cluster_centers,
            np_center = None if len(cluster_centers) > 1 else cluster_centers[0],
        )
        # This is the key metric for us to define optimal number of centers, by looking at the +/- change
        # when cluster centers increase/decrease
        centers_distance_median = np.median(distances_cluster_centers)
        self.logger.debug('At n=' + str(n_centers) + ' centers median distance ' + str(centers_distance_median))

        # median radius of each cluster center to cluster points
        inner_radiuses = []
        # number of points inside each cluster
        inner_sizes = []
        for i in range(len(cluster_centers)):
            points_in_cluster = x[cluster_labels == i]
            # get distances of points with respect to center reference point
            if len(points_in_cluster) > 0:
                inner_distances = self.fit_utils.get_point_distances(
                    np_tensors = points_in_cluster,
                    np_center = cluster_centers[i],
                    metric = metric,
                )
                inner_rad = np.median(inner_distances)
                # self.logger.debug('Cluster #' + str(i) + ', radius = ' + str(radius))
                inner_radiuses.append(inner_rad)
            else:
                self.logger.warning('No points in cluster i=' + str(i))
                inner_radiuses.append(np.nan)
            inner_sizes.append(len(points_in_cluster))

        inner_radiuses = np.array(inner_radiuses)
        inner_sizes = np.array(inner_sizes)
        return {
            'centers_median': centers_distance_median,
            'inner_radiuses': inner_radiuses,
            'cluster_sizes': inner_sizes,
        }

    def kmeans(
            self,
            x: np.ndarray,
            n_centers: int,
            x_labels: list = None,
            # for fine-tuning already trained clusters, thus no need to start all over again
            # useful for clusters of more than 1,000,000 points for example, where starting
            # again means another half day of fit training
            start_centers: np.ndarray = None,
            km_iters = 100,
            converge_diff_thr = 0.00001,
            # pass through mode
            test_mode = False,
    ):
        assert x.ndim == 2
        if x_labels is None:
            x_labels = list(range(len(x)))

        if not test_mode:
            self.logger.info('Start running kmeans in test mode ' + str(test_mode))
            kmeans = KMeans(
                n_clusters = n_centers,
                init = 'k-means++' if start_centers is None else start_centers,
                max_iter = km_iters,
                n_init = 10,
                random_state = 0
            )
            kmeans.fit(x)
            fit_centers = kmeans.cluster_centers_
            fit_cluster_numbers = kmeans.labels_
            fit_total_iters = kmeans.n_iter_
            fit_inertia = kmeans.inertia_
        else:
            # Change the given n_centers from user
            n_centers = len(x)
            self.logger.info('Not running kmeans in test mode ' + str(test_mode))
            np_tmp_labels = np.array(x_labels)
            tmp_unique_labels = np.unique(np_tmp_labels).tolist()
            # self.logger.info('Unique labels for test mode ' + str(tmp_unique_labels))

            fit_centers = x
            fit_cluster_numbers = np.arange(len(x))
            self.logger.info('Fit cluster numbers in test mode ' + str(fit_cluster_numbers))
            fit_cluster_numbers = fit_cluster_numbers.tolist()
            fit_total_iters = 0
            # 0 distance between centers and points
            fit_inertia = 0
            self.logger.info('Test mode values for cluster, fit cluster numbers ' + str(fit_cluster_numbers))

        additional_info = self.derive_additional_cluster_info(
            x = x,
            n_centers = n_centers,
            cluster_centers = fit_centers,
            cluster_labels = fit_cluster_numbers,
            metric = 'euclid',
        )

        if x_labels is not None:
            cluster_label_to_labelsori = self.map_centers_to_original_labels(
                labels_original = x_labels,
                labels_cluster = list(fit_cluster_numbers),
            )
        else:
            cluster_label_to_labelsori = None

        self.logger.info(fit_cluster_numbers)
        cluster_numbers = list(fit_cluster_numbers)
        #
        # Приложения должно быть сохранить только "cluster_centers" и "cluster_labels"
        # Если входиться новые точки, не надо тренировать снова, но только продольжать трейнинг с
        # этих же "cluster_centers". Такой шаги:
        #   - вычислить метки кластера новых точек
        #   - продолжать трейнинг с прошлых центров и их меток кластеров
        #
        return {
            'total_iterations': fit_total_iters,
            'n_centers': n_centers,
            # Group the indexes in same cluster
            'clusters': [
                [idx for idx, clbl in enumerate(cluster_numbers) if clbl==i_center]
                for i_center in range(n_centers)
            ],
            'cluster_centers': fit_centers,
            # e.g. [2, 2, 2, 0, 0, 0, 1, 1, 1] из меток пользавателя ['a', 'a', 'a', 'b', 'b', 'b', 'c', 'c', 'd']
            'cluster_labels': cluster_numbers,
            # как связываются между новыми метками кластерами и исходными метками пользавателя. e.g. {
            #    2: {'a': 1.0, 'b': 0.0, 'c': 0.0, 'd': 0.0},
            #    0: {'b': 1.0, 'a': 0.0, 'c': 0.0, 'd': 0.0},
            #    1: {'c': 0.6666666666666666, 'd': 0.3333333333333333, 'a': 0.0, 'b': 0.0}
            # }
            'cluster_label_to_original_labels': cluster_label_to_labelsori,
            'centers_median': additional_info['centers_median'],
            'inner_radiuses': additional_info['inner_radiuses'],
            'cluster_sizes': additional_info['cluster_sizes'],
            'points_inertia': fit_inertia,
        }

    def kmeans_optimal(
            self,
            x: np.ndarray,
            x_labels: list = None,
            km_iters = 100,
            max_clusters = 100,
            min_clusters = 2,
            plot = False,
            # by default if 25% of the clusters are single point clusters, we quit
            thr_single_clusters = 0.25,
            estimate_min_max = False,
            # pass through mode
            test_mode = False,
    ):
        assert x.ndim == 2

        if estimate_min_max:
            min_clusters, max_clusters = self.estimate_min_max_clusters(n=len(x))

        if test_mode:
            min_clusters = len(x)
            max_clusters = len(x)
            start_centers = x
            self.logger.info(
                'Test mode set to ' + str(test_mode) + ', overwriting min/max clusters to ' + str(min_clusters)
                + '/' + str(max_clusters) + ', and start centers to be just the points itself.'
            )
        else:
            start_centers = None

        # do a Monte-carlo
        # distances = self.fit_utils.get_point_distances_mc(np_tensors=self.text_encoded, iters=10000)
        # median_point_dist = np.median( distances )

        cluster_sets = {}
        max_n = min_clusters
        for n_centers in range(min_clusters, min(max_clusters+1, len(x)+1), 1):
            cluster_res = self.kmeans(
                x = x,
                x_labels = x_labels,
                n_centers = n_centers,
                km_iters = km_iters,
                start_centers = start_centers,
                test_mode = test_mode,
            )
            cluster_sets[n_centers] = cluster_res
            max_n = n_centers
            inner_sizes = cluster_res['cluster_sizes']

            # Check if clusters are becoming too sparse (1-member clusters too many)
            count_1_point_clusters = len(inner_sizes[inner_sizes==1])
            if count_1_point_clusters / len(inner_sizes) > thr_single_clusters:
                self.logger.info(
                    'Break at n centers = ' + str(n_centers) + ' with more than ' + str(100*thr_single_clusters)
                    + '% single point clusters, total single point clusters = '
                    + str(count_1_point_clusters)
                )
                break

        cs = cluster_sets
        val_cm = np.array([cs[i]['centers_median'] for i in range(max_n+1) if i>=min_clusters])
        #
        # Heuristic method using "pigeonhole principle". If n is the optimal number of clusters, then adding
        # one more will need to "crowd" the new center among the existing n centers. Thus center median should
        # reduce, or gradient calculated below is positive
        #
        grad_cm = [
            cs[i]['centers_median']-cs[i+1]['centers_median'] for i in range(max_n+1)
            if ((i >= min_clusters) and (i < max_n))
        ]
        grad_cm.append(0.)
        is_local_max_cm = [1*(x>0.00001) for x in grad_cm]
        if plot:
            self.logger.debug('Value CM: ' + str(val_cm))
            self.logger.debug('Gradient CM: ' + str(grad_cm))
            self.logger.debug('Is local max: ' + str(is_local_max_cm))
            mplt.plot(val_cm, linestyle='dotted')
            mplt.plot(is_local_max_cm)
            mplt.show()

        count_turning_point = 0
        for n_ctr, cluster_info in cluster_sets.items():
            is_turning_point = is_local_max_cm[n_ctr-min_clusters]
            count_turning_point += 1*is_turning_point

            cluster_info['count_turning_point'] = count_turning_point
            cluster_info['is_local_max_centers_median'] = is_turning_point
            cluster_info['gradient'] = grad_cm[n_ctr-min_clusters]

            if cluster_info['is_local_max_centers_median']:
                self.logger.info(
                    'Decrease of median distance of cluster centers at n_centers=' + str(n_ctr)
                    + ', from median distance ' + str(val_cm[n_ctr-min_clusters])
                    + ' to ' + str(val_cm[n_ctr+1-min_clusters])
                )

        final_clusters = [
            x for n,x in cluster_sets.items() if cluster_sets[n]['is_local_max_centers_median']
        ]
        if len(final_clusters) == 0:
            # There was no turning point, thus take the last one
            final_clusters = [x for n,x in cluster_sets.items() if (n==max_n)]
        return final_clusters

    def get_cluster_label_for_arbitrary_x(
            self,
            x_arbitrary: np.ndarray,
            cluster_centers: np.ndarray,
    ):
        assert x_arbitrary.ndim == cluster_centers.ndim, \
            'Different ndims ' + str(x_arbitrary.shape) + ', ' + str(cluster_centers.shape)
        top_pos, _ = self.tensor_utils.dot_sim(
            x = x_arbitrary,
            ref = cluster_centers,
            return_tensors = 'np',
        )
        # Just index 0 of every row
        return top_pos[:, 0]

    # Unlike PCA, cluster algorithm will destroy the concept of the original labels.
    # Thus we do a mapping back to the original labels using statistics of cluster centers.
    # For example, given the following labels & cluster numbers
    #      ['a', 'a', 'a', 'b', 'b', 'b', 'c', 'c', 'c'],
    #      [  0,   0,   1,   0,   1,   1,   0,   2,   2],
    # Probability map now looks like this
    #       {
    #          0: {'a': 0.5,   'b': 0.25,  'c': 0.25},
    #          1: {'a': 0.333, 'b': 0.666, 'c': 0.0 },
    #          2: {'a': 0.0,   'b': 0.0,   'c': 1.0 }
    #       }
    def map_centers_to_original_labels(
            self,
            labels_original: list,
            labels_cluster: list,
    ):
        map = ClusterMetrics(logger=self.logger).map_cluster_labels_to_original_labels(
            point_labels = labels_original,
            point_cluster_numbers = labels_cluster,
        )
        return map


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    clstr = Cluster()
    x = np.array([
        [5, 1, 1], [8, 2, 1], [6, 0, 2],
        [1, 5, 1], [2, 7, 1], [0, 6, 2],
        [1, 1, 5], [2, 1, 8], [0, 2, 6],
    ])
    labels = ['a', 'a', 'a', 'b', 'b', 'b', 'c', 'c', 'd']
    res = clstr.kmeans_optimal(x=x, x_labels=labels, estimate_min_max=True)
    for cluster_info in res:
        print('  Cluster ' + str(cluster_info['n_centers']))
        [print('    ' + str(k) + ': ' + str(v)) for k,v in cluster_info.items()]
        print('   Cluster map: ' + str())

    n_centers = res[0]['n_centers']
    c_centers = res[0]['cluster_centers']
    c_labels = res[0]['cluster_labels']
    # Fine Tune instead of retrain
    x_new = np.array([[7, 1, 1], [1, 0, 7]])
    # calculate label
    x_new_labels = clstr.get_cluster_label_for_arbitrary_x(
        x_arbitrary = x_new,
        cluster_centers = c_centers,
    )
    print('New cluster labels ' + str(x_new_labels))
    res = clstr.kmeans(
        x = np.append(x, x_new, axis=0),
        n_centers = n_centers,
        x_labels = list(c_labels) + list(x_new_labels),
        start_centers = c_centers,
    )
    print('  Cluster ' + str(res['n_centers']))
    [print('    ' + str(k) + ': ' + str(v)) for k,v in res.items()]
    print('   Cluster map: ' + str())
    exit(0)



import logging
import re
import os


class StringVar:

    PAT_VARNAME = r'[a-zA-Z._:\-]+'

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def replace_vars(
            self,
            value,
            # any function that takes the first argument as the variable name, returns value of variable
            get_var_value_func,
            # minimize use of stressful regex for user, just simple string
            var_string_start = '${',
            var_string_end   = '}',
            # avoid using regex if possible, it is over over-complicated & hard to maintain as always
            use_regex = False,
            # only required if use_regex = True
            varname_pattern  = PAT_VARNAME,
    ):
        if use_regex:
            # Put a bracket on every character
            pat_front = ''.join(['['+c+']' for c in var_string_start])
            pat_back = ''.join(['['+c+']' for c in var_string_end])
            # Final pattern to detect
            pattern = pat_front + r'(' + varname_pattern + ')' + pat_back
            self.logger.debug(
                'Using this pattern "' + str(pattern) + '" for value "' + str(value) + '"'
            )
        else:
            pattern = None

        # Index required if not using regex
        i = 0
        # Keep track of how many variables in the value
        count_var = 0
        # start/end index of variable found
        start, end = 0, 0
        # If not using regex
        mode_look_for_start_var = True

        while True:
            if i > len(value):
                break
            if not use_regex:
                try:
                    if mode_look_for_start_var:
                        if value[i:(i + len(var_string_start))] == var_string_start:
                            # Found start of var
                            mode_look_for_start_var = False
                            # include start variable brackets/etc
                            start = i
                            # fast forward to look for variable name
                            i += len(var_string_start)
                        else:
                            i += 1
                        continue

                    # If arrive here means we are looking for variable ending

                    if value[i:(i + len(var_string_end))] == var_string_end:
                        # Found end of var
                        mode_look_for_start_var = True
                        count_var += 1
                        # include end variable brackets/etc
                        end = i + len(var_string_end)
                        # fast forward to look for new variable
                        i = end + 1
                    else:
                        # keep looking for variable ending
                        i = i + 1
                        continue
                except:
                    # no more start or end to look for
                    break
            else:
                # re.match() will not work with this pattern, only re.search() or re.finditer() will work
                m = re.search(pattern=pattern, string=value)
                if not m:
                    self.logger.debug(
                        'No more variable pattern matches at iteration #' + str(i)
                        + ' for string value "' + str(value) + '"'
                    )
                    break
                count_var += 1
                self.logger.debug(
                    'Found variable #' + str(count_var) + ' from value "' + str(value) + '": ' + str(m)
                )
                start, end = m.span()[0], m.span()[1]

            # Take into account start characters (e.g. "${") and ending characters (e.g. "}")
            start_actual, end_actual = start + len(var_string_start), end - len(var_string_end)

            if start_actual >= end_actual:
                self.logger.error(
                    'Incorrect start ' + str(start_actual) + ' <= end ' + str(end_actual)
                    + ', regex pattern "' + str(pattern)
                    + '", for string "' + str(value) + '"'
                )
                break

            # Maintain simplicity, try not to use regex
            varname = value[start_actual:end_actual]
            # varname_with_brackets = value[start:end]
            # varname = re.sub(pattern=pattern, repl='\\1', string=varname_with_brackets)

            var_value = get_var_value_func(varname)

            self.logger.debug(
                'Var name "' + str(varname) + '" variable value = "' + str(var_value) + '"'
            )

            value = value[:start] + str(var_value) + value[end:]
            self.logger.debug(
                'Updated Value now after replacing variable #' + str(count_var) + ' "' + str(value) + '"'
            )
        return value


class StringVarUnitTest:
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self):
        sv = StringVar()
        os.environ['XXX'] = 'abc888'

        def example_get_var_value(*args):
            return {
                'FOLDER':       '/usr/local',
                'SECT::FOLDER': '/usr/share',
                'SECRET':       'abc123',
                'ENV_PASSWORD': os.environ['XXX'],
            }[args[0]]

        for string, var_str_start, var_str_end, expected_val in (
                ('1 ${FOLDER} xx',          '${',    '}', '1 /usr/local xx'),
                ('2 ${SECT::FOLDER} xx',    '${',    '}', '2 /usr/share xx'),
                ('3 $ENV{ENV_PASSWORD} xx', '$ENV{', '}', '3 abc888 xx'),
        ):
            for use_re in (True, False,):
                val = sv.replace_vars(
                    value = string,
                    var_string_start = var_str_start,
                    var_string_end = var_str_end,
                    get_var_value_func = example_get_var_value,
                    use_regex = use_re,
                    varname_pattern = StringVar.PAT_VARNAME if use_re else None,
                )
                assert val == expected_val,\
                    'Use regex "' + str(use_re) + '" Value get "' + str(val) + '" not equal to expected value "' \
                    + str(expected_val) + '"'
        print('ALL TESTS PASSED OK')
        return


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    StringVarUnitTest().test()
    exit(0)



# -*- coding: utf-8 -*-
import logging
import json
import numpy as np
from base64 import b64encode, b64decode
from fitxf.utils import Logging


class Base64:

    B64_ALLOWED_CHARSET = [chr(x) for x in range(ord('A'), ord('Z')+1)] + \
                          [chr(x) for x in range(ord('a'), ord('z')+1)] + \
                          [chr(x) for x in range(ord('0'), ord('9')+1)] + \
                          ['+', '/', '=']

    def __init__(
            self,
            text_encoding = 'utf-8',
            logger = None,
    ):
        self.text_encoding = text_encoding
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def encode(
            self,
            b,
    ) -> str:
        b = bytes(b, self.text_encoding) if type(b) is str else b
        assert type(b) is bytes
        e_bytes = b64encode(s=b)
        b64_str = e_bytes.decode(encoding=self.text_encoding)
        assert type(b64_str) is str
        return b64_str

    def decode(
            self,
            s: str,
    ) -> bytes:
        d_bytes = b64decode(s.encode(encoding=self.text_encoding))
        assert type(d_bytes) is bytes
        return d_bytes

    #
    # Note: If a string consists only of base64 characters [a-zA-Z0-9+/=], and have length
    # 0 (mod 3) then it will be wrongly recognized as a base64 string.
    #
    def is_base_64_string(
            self,
            s,
    ):
        s = str(s)
        zero_mod3 = (len(s) % 3 == 0)
        is_b64_charset = True
        for char in s:
            if char not in self.B64_ALLOWED_CHARSET:
                is_b64_charset = False
                break

        return zero_mod3 & is_b64_charset

    def encode_numpy_array_to_base64_string_multidim(
            self,
            x: np.ndarray,
            # can be string like 'float64', or numpy dtype such as np.float64
            data_type = np.float64,
    ) -> str:
        x_shape = x.shape
        b64_str_flattenned = self.encode_numpy_array_to_base64_string(x=x, data_type=data_type)
        return json.dumps({'shape': list(x_shape), 'b64_str': b64_str_flattenned})

    def decode_base64_string_to_numpy_array_multidim(
            self,
            string: str,
            data_type = np.float64,
    ) -> np.ndarray:
        data = json.loads(string)
        x_shape = data['shape']
        b64_str_flattenned = data['b64_str']
        x_flattenned = self.decode_base64_string_to_numpy_array(s64=b64_str_flattenned, data_type=data_type)
        return x_flattenned.reshape(x_shape)

    # Warning: Encoding numpy array to bytes will flatten it to 1-dimensional
    def encode_numpy_array_to_base64_string(
            self,
            x: np.ndarray,
            # can be string like 'float64', or numpy dtype such as np.float64
            data_type = np.float64,
    ) -> str:
        if x.ndim > 1:
            self.logger.warning(
                'Encoding to base 64 will flatten numpy ndim=' + str(x.ndim) + ' to ndim=1.'
            )
        # Step 1: Convert numpy to float64 bytes
        x_tp = x.astype(dtype=data_type)
        x_tp_bytes = x_tp.tobytes()
        # Step 2 & 3: then convert to base 64 bytes, then to base 64 utf-8 string
        return b64encode(x_tp_bytes).decode('utf-8')

    def decode_base64_string_to_numpy_array(
            self,
            s64: str,
            data_type = np.float64,
    ) -> np.ndarray:
        # Step 1: Convert base 64 string to base 64 bytes
        s64_b = s64.encode('utf-8')
        # Step 2: Convert base 64 bytes to actual bytes
        actual_bytes = b64decode(s64_b)
        # Step 3: Finally convert to numpy array from base 64 bytes
        return np.frombuffer(actual_bytes, dtype=data_type)


class Base64UnitTest():

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self):
        logger = Logging.get_default_logger(log_level=logging.DEBUG, propagate=False)
        b64 = Base64(logger=logger)
        invalid_b64_strings = ['no b64', '1231234']
        for s in invalid_b64_strings:
            assert b64.is_base_64_string(s=s) is False, 'String "' + str(s) + '" should not be a base64 string.'

        #
        # Test string encoding/decoding
        #
        tests = (
            ('string to be encoded', 'c3RyaW5nIHRvIGJlIGVuY29kZWQ='),
            ('한국 인터넷 라디오 방송국', '7ZWc6rWtIOyduO2EsOuEtyDrnbzrlJTsmKQg67Cp7Iah6rWt'),
        )
        for s_or_np, expected_b64_str in tests:
            e_str = b64.encode(b=s_or_np)
            d_bytes = b64.decode(s=e_str)
            d_str_or_np = d_bytes.decode(encoding=b64.text_encoding)
            self.logger.info(
                'Original object <<' + str(s_or_np) + '>> encoded to <<' + str(e_str)
                + '>>, decoded back as <<' + str(d_str_or_np) + '>>'
            )
            assert e_str == expected_b64_str, 'Encoded b64 "' + str(e_str) + '" not "' + str(expected_b64_str) + '"'
            assert d_str_or_np == s_or_np, 'Decoded b64 "' + str(d_str_or_np) + '" not "' + str(s_or_np) + '"'

        #
        # Test numpy encoding/decoding of bytes to base 64
        #
        np_tests = (
            (np.array([1.23, 4.55, 7.43, 555.42]), 'rkfhehSu8z8zMzMzMzMSQLgehetRuB1Aj8L1KFxbgUA='),
            (np.array([[1.23, 4.55], [7.43, 555.42]]), 'rkfhehSu8z8zMzMzMzMSQLgehetRuB1Aj8L1KFxbgUA='),
        )
        for x, expected_b64_str in np_tests:
            e_str = b64.encode_numpy_array_to_base64_string(x=x, data_type=x.dtype)
            d_np = b64.decode_base64_string_to_numpy_array(s64=e_str, data_type=x.dtype)
            self.logger.info(
                'Original object <<' + str(x) + '>> encoded to <<' + str(e_str)
                + '>>, decoded back as <<' + str(d_np) + '>>'
            )
            assert e_str == expected_b64_str, \
                'Encoded b64 "' + str(e_str) + '" not "' + str(expected_b64_str) + '"'
            assert np.sum((d_np - x.flatten()) ** 2) < 0.0000000001, \
                'Decoded b64 "' + str(d_np) + '" not "' + str(x) + '"'

        # Test multidim
        np_tests = (
            (
                np.array([1.23, 4.55, 7.43, 555.42]),
                {'shape': [4], 'b64_str': 'rkfhehSu8z8zMzMzMzMSQLgehetRuB1Aj8L1KFxbgUA='},
            ),
            (
                np.array([[1.23, 4.55], [7.43, 555.42]]),
                {'shape': [2,2], 'b64_str': 'rkfhehSu8z8zMzMzMzMSQLgehetRuB1Aj8L1KFxbgUA='}
            ),
        )
        for x, expected_b64_json in np_tests:
            e_str = b64.encode_numpy_array_to_base64_string_multidim(x=x, data_type=x.dtype)
            d_np = b64.decode_base64_string_to_numpy_array_multidim(string=e_str, data_type=x.dtype)
            self.logger.info(
                'Original object <<' + str(x) + '>> encoded to <<' + str(e_str)
                + '>>, decoded back as <<' + str(d_np) + '>>'
            )
            assert json.loads(e_str) == expected_b64_json, \
                'Encoded b64 json "' + str(e_str) + '" not "' + str(expected_b64_json) + '"'
            assert np.sum((d_np - x) ** 2) < 0.0000000001, \
                'Decoded b64 "' + str(d_np) + '" not "' + str(x) + '"'

        print('B64 TESTS PASSED OK')
        return


if __name__ == '__main__':
    Base64UnitTest(logger=Logging.get_default_logger(log_level=logging.INFO, propagate=False)).test()
    exit(0)



import logging
import os
import threading
import random
import uuid
import time as t
import datetime as dt
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Profile import Profiling


#
# THEORY OF CLASH
#   Both Models below have been tested and is theoretically correct.
#
#   Model:
#     Let the frequency of checks for file lock availability be C per second
#     for each worker. As example let C=4.
#     Let the available lock "slots" available in 1s be S.
#     Since it takes about 0.3521 ms at 1% quantile, we assume the clash interval
#     at 0.35ms or about 3000-th of a second. If we assume the lock is held for
#     another period of time to do other things, we assume the lock is held for
#     10ms, or S=100.
#     Thus in 1s if S=100, there are only S=100 lock slots on offer.
#     If the number of threads competing is N>100, then some threads will not
#     gain a slot in that 1s window.
#     This means the probability of checking in any of the write slot is C/W
#     or about 1/100 (note it is not a Poisson).
#     This means in 1s, the probability of a thread not gaining a lock if in
#     total N threads are competing is given by the access slots max(0,N-S)
#     divided by total threads N
#            F(1) = P(Fail to obtain lock in 1s) = max(0,N-S)/N
#     Thus the probability of not obtaining lock in x seconds (assuming all N
#     threads are working forever, never stops) is
#            F(x) = F(1)/x
#
#   Empirical Model:
#     But for simplicity, we can use the following:
#     When N workers/processes/threads simultaneously access a resource, with
#     max wait time W, there is a probability of a worker never getting to access
#     this resource within this time W.
#     If the wait time W is doubled, the probability falls by half (not mathematically
#     correct actually).
#     If the number of threads N are doubled, the probability increases twice (also
#     not mathematically correct).
#     Thus
#          P(fail_to_obtain_lock) = k * N / W
#     The constant k depends on the machine, on a Mac Book Air, k = 1/250 = 0.005
#
class LockFile:

    N_RACE_CONDITIONS_FILE = 0

    #
    # If a lock file is older than 30 secs, remove forcefully.
    # No process should hold it for so long anyway.
    #
    FORCE_REMOVE_LOCKFILE_AGE_SECS_THRESHOLD = int(os.environ.get("FORCE_REMOVE_LOCKFILE_AGE_SECS_THRESHOLD", "30"))

    # For Mac Book Air
    K_CONSTANT_MAC_BOOK_AIR = 1 / 250
    # Slots available per second
    SLOTS_PER_SEC = 45

    def __init__(
            self,
            lock_file_path,
            logger = None,
    ):
        self.lock_file_path = lock_file_path
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def __wait_for_lock_file(
            self,
            max_wait_time_secs,
    ):
        total_sleep_time = 0.0

        while os.path.exists(self.lock_file_path):
            #
            # Check lock file time, if older than certain threshold, remove it forcefully.
            # It is unlikely a process has had it for so long, or that process may have died
            # already.
            #
            # Check if file time is newer
            try:
                ftime = dt.datetime.fromtimestamp( os.path.getmtime(self.lock_file_path) )
                lockfile_age_secs = Profiling(logger=self.logger).get_time_dif_secs(
                    start=ftime, stop=dt.datetime.now(), decimals=4
                )
            except Exception as ex_check_time:
                lockfile_age_secs = -1
                self.logger.warning(
                    'Exception checking time for lock file "' + str(self.lock_file_path)
                    + '", possible already released by another process. Exception: ' + str(ex_check_time)
                )

            if lockfile_age_secs > self.FORCE_REMOVE_LOCKFILE_AGE_SECS_THRESHOLD:
                self.logger.critical(
                    'Trying to forcefully remove lock file "' + str(self.lock_file_path)
                    + '". Already ' + str(lockfile_age_secs) + ' seconds old.'
                )
                if self.release_file_cache_lock():
                    self.logger.warning(
                        'Success. Forcefully removed lock file "' + str(self.lock_file_path)
                        + '". Already ' + str(lockfile_age_secs) + ' seconds old.'
                    )
                else:
                    self.logger.critical(
                        'Error. Failed to forcefully remove lock file "' + str(self.lock_file_path)
                        + '". Already ' + str(lockfile_age_secs) + ' seconds old.'
                    )

            self.logger.info(
                'Waiting for file lock "' + str(self.lock_file_path)
                + '", ' + str(round(total_sleep_time,2))
                + 's. Lock file age = ' + str(lockfile_age_secs) + 's.'
            )
            sleep_time = random.uniform(0.1,0.5)
            t.sleep(sleep_time)
            total_sleep_time += sleep_time
            if total_sleep_time > max_wait_time_secs:
                self.logger.warning(
                    'Wait fail for file lock "' + str(self.lock_file_path) + '" after '
                    + str(round(total_sleep_time,2)) + ' secs!!'
                )
                return False
        return True

    def acquire_file_cache_lock(
            self,
            # At 10s max wait time, this means the probability of not obtaining lock if
            # 10 processes simultaneously wants to access it is roughly (1/250)*10/10 = 0.4%
            max_wait_time_secs = 10.0,
    ):
        #
        # At this point there could be many competing workers/threads waiting for it.
        # And since they can be cross process, means no point using any mutex locks.
        #
        wait_time_per_round = 0.5
        self.logger.debug('Wait time per round ' + str(round(wait_time_per_round,2)))
        random_val = wait_time_per_round / 10
        total_wait_time = 0
        round_count = 0
        while True:
            if total_wait_time >= max_wait_time_secs:
                self.logger.critical(
                    'Round ' + str(round_count)
                    + '. Failed to get lock ~' + str(total_wait_time) + 's to file "'
                    + str(self.lock_file_path) + '"! Very likely process is being bombarded with too many requests.'
                )
                return False
            round_count += 1

            # Rough estimation without the random value
            total_wait_time += wait_time_per_round

            if not self.__wait_for_lock_file(
                    max_wait_time_secs = random.uniform(
                        wait_time_per_round-random_val,
                        wait_time_per_round+random_val
                    )
            ):
                self.logger.info(
                    'Round ' + str(round_count) + ' fail to get lock to file "' + str(self.lock_file_path) + '".'
                )
                continue
            else:
                self.logger.debug('Lock file "' + str(self.lock_file_path) + '" ok, no longer found.')

            try:
                f = open(file=self.lock_file_path, mode='w')
                timestamp = dt.datetime.now()
                random_string = uuid.uuid4().hex + ' ' + str(timestamp) + ' ' + str(threading.get_ident())
                self.logger.debug(
                    'Write random string "' + str(random_string) + '" to lock file "' + str(self.lock_file_path) + '".'
                )
                f.write(random_string)
                f.close()

                #
                # If many processes competing to obtain lock, means many processes will arrive here simultaneously.
                # Read back, as there might be another worker/thread that reached this point and wrote
                # something to it also. This can handle cross process clashes, unlike memory locks.
                #
                # The time stats for writing a lock file (test function below) is the following in milliseconds:
                #   {'average': 0.6627,
                #   'quantile': {'0.001%': 0.3521, '1.0%': 0.369, '10.0%': 0.407,
                #   '25.0%': 0.447, '50.0%': 0.523, '75.0%': 0.67, '90.0%': 0.898, '99.0%': 1.966,
                #   '99.999%': 176.8324}}
                # This means by sleeping between 10ms to 15ms is more than enough to cover at least 99% (1.966ms)
                # of the cases.
                # Meaning to say if there are race conditions, after the sleep, all competitors would have
                # already written to the file, thus when reading back, only 1 competitor will see it correctly.
                t.sleep(0.01+random.uniform(0.0, 0.005))
                self.logger.debug(
                    'Check random string "' + str(random_string) + '" from lock file "' + str(self.lock_file_path) + '".'
                )
                f = open(file=self.lock_file_path, mode='r')
                read_back_string = f.read()
                f.close()
                if (read_back_string == random_string):
                    self.logger.debug('Read back random string "' + str(read_back_string) + '" ok.')
                    return True
                else:
                    self.N_RACE_CONDITIONS_FILE += 1
                    self.logger.warning(
                        'File Race condition ' + str(self.N_RACE_CONDITIONS_FILE)
                        + '! Round ' + str(round_count)
                        + '. Failed verify lock file with random string "'
                        + str(random_string) + '", got instead "' + str(read_back_string) + '".'
                    )
                    continue
            except Exception as ex_file:
                self.logger.error(
                    'Round ' + str(round_count) + '. Error lock file "' + str(self.lock_file_path)
                    + '": ' + str(ex_file)
                )
                continue
            finally:
                pass

        return False

    def release_file_cache_lock(
            self,
    ):
        if not os.path.isfile(self.lock_file_path):
            self.logger.warning('No lock file "' + str(self.lock_file_path) + '" to release!!')
            return True

        try:
            #
            # It is not possible for multiple processes to want to remove the lock
            # simultaneously since at any one time there should only be 1 process
            # having the lock.
            # So means there is no need to use mutexes.
            #
            os.remove(self.lock_file_path)
            self.logger.info('Lock file "' + str(self.lock_file_path) + '" removed/released successfully.')
            return True
        except Exception as ex:
            self.logger.critical('Unable to remove lock file "' + str(self.lock_file_path) + '": ' + str(ex))
            return False

    def get_time_stats_to_create_lock_file(
            self,
            n_rounds = 1000
    ):
        x = []

        for i in range(n_rounds):
            a = Profiling(logger=self.logger).start()
            f = open(file=self.lock_file_path, mode='w')
            timestamp = dt.datetime.now()
            random_string = uuid.uuid4().hex + ' ' + str(timestamp) + ' ' + str(threading.get_ident())
            self.logger.debug(
                'Write random string "' + str(random_string) + '" to lock file "' + str(self.lock_file_path) + '".'
            )
            f.write(random_string)
            f.close()
            time_taken_millisecs = Profiling.get_time_dif_secs(start=a, stop=Profiling.stop(), decimals=6) * 1000
            x.append(time_taken_millisecs)
            print(str(i+1) + ': ' + str(time_taken_millisecs) + 'ms')

        import numpy as np
        x_np = np.array(x)
        quantiles = {}
        for q in [0.001, 1.0, 10.0, 25.0, 50.0, 75.0, 90.0, 99.0, 99.999]:
            quantiles[str(q)+'%'] = round(np.quantile(x_np, q/100), 4)

        return {
            'average': round(x_np.mean(),4),
            'quantile': quantiles
        }


class LoadTestLockFile:

    def __init__(
            self,
            lock_file_path,
            max_wait_time_secs,
            n_threads,
            count_to,
            logger = None,
    ):
        self.lock_file_path = lock_file_path
        self.max_wait_time_secs = max_wait_time_secs
        self.n_threads = n_threads
        self.count_to = count_to
        self.logger = logger if logger is not None else logging.getLogger()

        self.x = 0
        self.n_failed = 0
        self.__mutex = threading.Lock()
        self.lf_obj = LockFile(lock_file_path=self.lock_file_path, logger=self.logger)
        return

    def __safe_incr(self, x_inc=0, n_failed_inc=0):
        try:
            self.__mutex.acquire()
            self.x += x_inc
            self.n_failed += n_failed_inc
        finally:
            self.__mutex.release()

    def target(
            self,
            thread_name,
            count,
    ):
        for i in range(count):
            if self.lf_obj.acquire_file_cache_lock(
                    max_wait_time_secs = self.max_wait_time_secs,
            ):
                self.__safe_incr(x_inc=1)
                print('x = ' + str(self.x) + ', thread ' + str(thread_name))
                self.lf_obj.release_file_cache_lock()
            else:
                self.__safe_incr(n_failed_inc=1)
                print(
                    '***** ' + str(self.n_failed) + '. Failed to obtain lock: ' + str(self.x)
                )
        print('***** THREAD ' + str(thread_name) + ' DONE ' + str(count) + ' COUNTS')

    def run(self):
        threads_list = []
        n_sum = 0
        for i in range(self.n_threads):
            n_sum += self.count_to
            thr = threading.Thread(target=self.target, args=[str(i), self.count_to])
            threads_list.append(thr)
            print(str(i) + '. New thread "' + str() + '" count ' + str(self.count_to))
            print('Starting thread ' + str(i))
            thr.start()

        for thr in threads_list:
            thr.join()

        print('********* THREADS N=' + str(self.n_threads) + ', WAIT=' + str(self.max_wait_time_secs) + 's.')
        print('********* TOTAL SHOULD GET ' + str(n_sum) + '. Failed Counts = ' + str(self.n_failed))
        print('********* TOTAL COUNT SHOULD BE = ' + str(n_sum - self.n_failed))
        print('********* TOTAL RACE CONDITIONS FILE = ' + str(self.lf_obj.N_RACE_CONDITIONS_FILE))
        print('********* PROBABILITY OF FAILED LOCKS = ' + str(round(self.n_failed / n_sum, 4)))
        N = self.n_threads
        x = self.max_wait_time_secs
        F_1 = max(0, N-LockFile.SLOTS_PER_SEC) / N
        print('********* THEO PROBABILITY OF FAILED LOCKS (Theo Model) = '
              + str(round(F_1/x, 4)))
        print('********* THEO PROBABILITY OF FAILED LOCKS (Empirical Model) = '
              + str(round(LockFile.K_CONSTANT_MAC_BOOK_AIR * self.n_threads / self.max_wait_time_secs, 4)))


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    lfp = 'lockfile.test.lock'

    lf = LockFile(lock_file_path=lfp, logger=lgr)
    lf.acquire_file_cache_lock()
    lf.release_file_cache_lock()

    test = 'load test lock file'

    if test == 'lock file stats':
        print('Average time to create lock file = ')
        print(LockFile.get_time_stats_to_create_lock_file(n_rounds=10000))
        exit(0)

    if test == 'load test lock file':
        LoadTestLockFile(
            lock_file_path = lfp,
            # From trial and error, for 100 simultaneous threads, each counting to 50,
            # waiting for max 10 secs, the probability of failed lock is about 100/5000
            # If the wait time W is doubled, the probability falls by half.
            # If the number of threads N are doubled, the probability increases twice.
            # Thus P(fail_lock) = k * N / W
            # The constant k depends on the machine, on a Mac Book Air, k = 1/200 = 0.005
            max_wait_time_secs = 8.8,
            n_threads = 5,
            # The probability of failed lock does not depend on this, this is just sampling
            count_to = 10,
            logger = lgr,
        ).run()

    if test == '':
        lf = LockFile(logger=lgr)
        res = lf.acquire_file_cache_lock(
            max_wait_time_secs = 30
        )
        print('Lock obtained = ' + str(res))
        res = lf.release_file_cache_lock()
        print('Lock released = ' + str(res))

        res = lf.acquire_file_cache_lock(
            max_wait_time_secs = 2.2
        )
        print('Lock obtained = ' + str(res))



import logging
import networkx as nx
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Pandas import Pandas


class GraphUtils:

    def __init__(
            self,
            logger = None,
    ):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def create_multi_graph(
            self,
            # expect compulsory keys in dict "key", "u", "v", "weight"
            edges: list[dict],
            # compulsory columns in each dict
            col_u = 'u',
            col_v = 'v',
            col_key = 'key',
            col_weight = 'weight',
            directed = False,
    ) -> nx.Graph:
        self.logger.debug('Directed graph = ' + str(directed) + '.Edges to create graph: ' + str(edges))
        multi_g = nx.MultiDiGraph() if directed else nx.MultiGraph()

        for i, edge_rec in enumerate(edges):
            self.logger.debug('#' + str(i) + ': ' + str(edge_rec))
            u = edge_rec[col_u]
            v = edge_rec[col_v]
            key = edge_rec[col_key]
            weight = edge_rec[col_weight]
            other_params = {k: v for k, v in edge_rec.items() if k not in [col_u, col_v, col_key, col_weight]}
            edge_key = (u, v, key)
            if multi_g.edges.get(edge_key) is not None:
                self.logger.warning(str(i) + '. Edge exists ' + str(edge_key) + ': ' + str(multi_g.edges.get(edge_key)))
            else:
                self.logger.debug(str(i) + '. New edge ' + str(edge_key))
            # There will be no duplicate edges, just overwritten by the last one
            multi_g.add_edge(
                # For type nx.Graph, order of u, v does not matter, searching for the edge (u, v)
                # or (v, u) will return the same thing
                key = key,
                u_for_edge = u,
                v_for_edge = v,
                # User info
                u = u,
                v = v,
                weight = weight,
                params = other_params,
            )
            self.logger.debug(
                'Check edge symmetry, key ' + str((u, v)) + ' retrieved from graph ' + str(multi_g.edges.get(edge_key))
                + ', opposite key ' + str((v, u)) + ' retrieved as ' + str(multi_g.edges.get(edge_key))
            )
        return multi_g

    def get_neighbors(self, G: nx.Graph, node: str):
        return nx.neighbors(G=G, n=node)

    def __get_leg_obj(self, src_tgt, leg_count, i_leg, leg_key, leg_a, leg_b, leg_weight):
        legs_in_start_end = 1*(leg_a in src_tgt) + 1*(leg_b in src_tgt)
        return {
            'src_tgt': src_tgt,
            'leg_total': leg_count,
            'leg_number': i_leg,
            'leg_key': leg_key,
            'leg_a': leg_a,
            'leg_b': leg_b,
            'legs_in_srctgt': legs_in_start_end,
            'leg_weight': leg_weight,
            'leg_weight_proportion': None,
        }

    def get_paths(
            self,
            G: nx.Graph,
            source,
            target,
            # permitted values "simple", "dijkstra", "shortest"
            method = 'dijkstra',
            # only applicable for "simple" path method
            agg_weight_by: str = 'min',
    ) -> list[dict]:
        assert method in ['simple', 'dijkstra', 'shortest']
        if method == 'dijkstra':
            agg_weight_by = 'min'
        source_target = (source, target)

        func = nx.dijkstra_path if method in ['dijkstra'] else (
            nx.shortest_path if method in ['shortest'] else nx.shortest_simple_paths
        )
        if method in ['simple']:
            # "simple" method cannot work with multigraph
            G__ = self.convert_multigraph_to_simple_graph(G=G, agg_weight_by=agg_weight_by)
            self.logger.info('Converted graph to non-multigraph for get paths method "' + str(method) + '"')
        else:
            G__ = G
        try:
            nodes_traversed_paths = func(
                G = G__,
                source = source,
                target = target,
            )
            if method == 'simple':
                nodes_traversed_paths = list(nodes_traversed_paths)
            else:
                # for "dijkstra", "shortest" will only return 1 path, we convert to list
                nodes_traversed_paths = [nodes_traversed_paths]
            self.logger.debug('Nodes traversed path "' + str(method) + '": ' + str(nodes_traversed_paths))
        except Exception as ex_no_path:
            self.logger.error(
                'Path "' + str(method) + '" from "' + str(source) + '" --> "' + str(target) + '": ' + str(ex_no_path)
            )
            return []

        paths_by_method = []
        for nodes_traversed_path in nodes_traversed_paths:
            best_legs = []
            nodes_traversed_weight = 0
            for i_leg in range(1, len(nodes_traversed_path)):
                leg_a = nodes_traversed_path[i_leg - 1]
                leg_b = nodes_traversed_path[i_leg]
                leg_ab = (leg_a, leg_b)
                self.logger.debug('Method "' + str(method) + '" checking leg ' + str(leg_ab))
                # if multiple will be by dictionary key: edge, e.g.
                # {
                #    'teleport': {'u': 'Tokyo', 'v': 'Beijing', 'weight': 2},
                #    'plane': {'u': 'Tokyo', 'v': 'Beijing', 'weight': 9}
                #  }
                ep = G.get_edge_data(u=leg_a, v=leg_b)
                self.logger.debug('For get path leg ' + str(leg_ab) + ', edge data ' + str(ep))
                # convert to convenient tuples instead of key: values
                ep_edges = [(k, d) for k, d in ep.items()]

                if agg_weight_by == 'min':
                    arg_best_weight = np.argmin([d['weight'] for k, d in ep_edges])
                else:
                    arg_best_weight = np.argmax([d['weight'] for k, d in ep_edges])

                best_key, best_edge = ep_edges[arg_best_weight]
                best_leg = self.__get_leg_obj(
                    src_tgt = source_target,
                    leg_count = len(nodes_traversed_path) - 1,
                    i_leg = i_leg,
                    leg_a = leg_a,
                    leg_b = leg_b,
                    leg_weight = best_edge['weight'],
                    leg_key = best_key,
                )
                nodes_traversed_weight += best_edge['weight']
                best_legs.append(best_leg)
            # Calculate weight proportion
            for lg in best_legs:
                lg['leg_weight_proportion'] = lg['leg_weight'] / nodes_traversed_weight

            paths_by_method.append({
                'path': nodes_traversed_path,
                'legs': best_legs,
                'weight_total': nodes_traversed_weight,
            })
        self.logger.debug(
            'Paths by method ' + str(paths_by_method)
        )
        return paths_by_method

    def __helper_convert_to_edge_path_dict(
            self,
            paths_dict: dict,
    ) -> dict:
        edge_path_dict = {}
        for start in paths_dict.keys():
            d_dest = paths_dict[start]
            [
                self.logger.debug(str(start) + '-->' + str(dest) + ':' + str(path))
                for dest, path in d_dest.items() if start != dest
            ]
            for dest, path in d_dest.items():
                if start != dest:
                    edge_path_dict[(start, dest)] = path
        return edge_path_dict

    def get_dijkstra_path_all_pairs(
            self,
            G: nx.Graph,
    ) -> dict:
        sp = dict(nx.all_pairs_dijkstra_path(G))
        return self.__helper_convert_to_edge_path_dict(paths_dict=sp)

    def get_shortest_path_all_pairs(
            self,
            G: nx.Graph,
    ) -> dict:
        sp = dict(nx.all_pairs_shortest_path(G))
        return self.__helper_convert_to_edge_path_dict(paths_dict=sp)

    # Given a set of edges, we find the paths traversed
    # TODO not yet optimized mathematically
    def search_top_keys_for_edges(
            self,
            query_edges: list[dict],
            ref_multigraph: nx.Graph,
            # permitted values "simple", "dijkstra", "shortest"
            path_method = 'dijkstra',
            # only applicable for "simple" path method
            path_agg_weight_by: str = 'min',
            query_col_u = 'u',
            query_col_v = 'v',
            # query_col_key = 'key',
            query_col_weight = 'weight',
    ):
        multi_graph = ref_multigraph
        self.logger.debug('Ref graph edges: ' + str(multi_graph.edges))
        self.logger.debug('Ref graph nodes: ' + str(multi_graph.nodes))
        if path_method == 'dijkstra':
            path_agg_weight_by = 'min'

        all_legs = []
        query_edges_best_paths = {}
        for i, conn in enumerate(query_edges):
            # for each query edge, find best legs
            self.logger.debug('For conn ' + str(conn))
            u = conn[query_col_u]
            v = conn[query_col_v]
            w_ref = conn.get(query_col_weight, 0)
            edge = (u, v)
            res = self.get_paths(
                G = multi_graph,
                source = u,
                target = v,
                method = path_method,
                agg_weight_by = path_agg_weight_by,
            )
            self.logger.debug(
                'Query edge #' + str(i) + ' method ' + str(path_method) + ', best paths for edge ' + str(edge)
                + ': ' + str(res)
            )
            if len(res) > 0:
                # if reference weight exists, take path with closest weight
                if path_agg_weight_by == 'min':
                    i_best = np.argmin([abs(d['weight_total'] - w_ref) for d in res])
                else:
                    i_best = np.argmax([abs(d['weight_total'] - w_ref) for d in res])
                self.logger.debug('Best path for method ' + str(path_method) + ': ' + str(res[i_best]))
                best_path_uv = res[i_best]['path']
                best_legs_uv = res[i_best]['legs']
                best_weight_total_uv = res[i_best]['weight_total']
            else:
                best_path_uv = None
                best_legs_uv = [self.__get_leg_obj(
                    src_tgt = edge,
                    leg_count = np.inf,
                    i_leg = None,
                    leg_a = None,
                    leg_b = None,
                    leg_weight = None,
                    leg_key = None,
                )]
                best_weight_total_uv = None
            self.logger.debug('Best path for ' + str((u, v)) + ': ' + str(best_path_uv))
            self.logger.info(
                'Conn #' + str(i) + ' for edge ' + str(edge) + ', best path: ' + str(best_path_uv)
                + ', best legs for ' + str((u, v)) + ': ' + str(best_legs_uv)
            )
            if best_legs_uv is not None:
                all_legs = all_legs + best_legs_uv
            query_edges_best_paths[(u, v)] = best_path_uv

        self.logger.info('Path shortest distances: ' + str(query_edges_best_paths))

        # Sort by shortest path to longest
        df_all_legs = pd.DataFrame.from_records(all_legs)
        #cond_na = df_all_legs['leg_weight'].isna()
        #df_all_legs['leg_weight'][cond_na] = np.inf if path_agg_weight_by == 'min' else 0.
        df_all_legs.sort_values(
            by = ['src_tgt', 'leg_total', 'leg_number', 'leg_a', 'leg_b', 'leg_weight'],
            ascending = True,
            inplace = True,
        )
        max_legs_uniq = np.unique(df_all_legs['leg_total']).tolist()
        max_legs_uniq.sort()
        self.logger.info(
            'Query-collections connections, max leg unique ' + str(max_legs_uniq)
            + ':\n' + str(df_all_legs)
        )

        #
        # Top keys by source-target
        #

        #
        # Top keys by weight proportion
        #
        keep_cols = ['src_tgt', 'leg_key', 'leg_weight', 'leg_weight_proportion']
        df_top_keys_by_agg_weight = df_all_legs[keep_cols].reset_index(drop=True)
        # MUST convert to numpy ndarray before multiply. pandas will fuck up with nans
        df_top_keys_by_agg_weight['__weight'] = np.array(df_top_keys_by_agg_weight['leg_weight']) \
                                                * np.array(df_top_keys_by_agg_weight['leg_weight_proportion'])
        self.logger.debug('Dataframe top keys by agg weight\n' + str(df_top_keys_by_agg_weight))
        keep_cols = ['src_tgt', 'leg_key', '__weight']
        df_top_keys_by_agg_weight = df_top_keys_by_agg_weight[keep_cols].groupby(
            by = ['src_tgt', 'leg_key'],
            as_index = False,
        ).sum()
        df_top_keys_by_agg_weight.sort_values(
            by=['__weight'], ascending=True if path_agg_weight_by=='min' else False, inplace=True
        )
        self.logger.info(
            'Top keys by aggregated weight (agg by "' + str(path_agg_weight_by) + '") sum\n'
            + str(df_top_keys_by_agg_weight)
        )
        top_keys_by_agg_weight = df_top_keys_by_agg_weight.to_dict(orient='records')
        for r in top_keys_by_agg_weight:
            r['__weight'] = round(r['__weight'], 3)

        #
        # Top keys by number of edges traversed
        #
        top_keys_by_number_of_edges = {}
        for leg_total in max_legs_uniq:
            condition = df_all_legs['leg_total'] == leg_total
            keys_uniq = list(set(df_all_legs[condition]['leg_key'].tolist()))
            keys_uniq.sort()
            # key is how many edges required
            top_keys_by_number_of_edges[leg_total] = keys_uniq
        self.logger.info('Top keys by number of edges: ' + str(top_keys_by_number_of_edges))

        # Indicators
        coverage = round(
            np.sum(
                [1 for v in query_edges_best_paths.values() if v is not None]
            ) / len(query_edges_best_paths.keys()), 3
        )
        self.logger.info(
            'Coverage = ' + str(coverage) + ', path most suitable distances ' + str(query_edges_best_paths)
        )

        return {
            'method': path_method,
            'path_agg_weight_by': path_agg_weight_by,
            'top_keys_by_aggregated_weight': top_keys_by_agg_weight,
            'top_keys_by_number_of_edges': top_keys_by_number_of_edges,
            'indicators': {
                'coverage': coverage,
                'best_paths': query_edges_best_paths,
            },
            'leg_details': df_all_legs,
        }

    def convert_multigraph_to_simple_graph(
            self,
            G: nx.Graph,
            agg_weight_by: str = 'min',
    ):
        if type(G) in [nx.MultiGraph, nx.MultiDiGraph]:
            # convert to non-multi graph to draw
            G_simple = nx.Graph()
            for edge in G.edges:
                u, v, key = edge
                e_data = G.get_edge_data(u=u, v=v)
                weights = [d['weight'] for key, d in e_data.items()]
                w = np.max(weights) if agg_weight_by == 'max' else np.min(weights)
                G_simple.add_edge(u_of_edge=u, v_of_edge=v, weight=w)
            self.logger.info(
                'Converted type "' + str(type(G)) + '" to type "' + str(G_simple) + '"'
            )
        else:
            G_simple = G
        return G_simple

    def draw_graph(
            self,
            G: nx.Graph,
            weight_large_thr: float = 0.5,
            # if multigraph, aggregate weight method
            agg_weight_by: str = 'min',
            draw_node_size: int = 100,
            draw_font_size:int = 16,
            draw_line_width: int = 4,
    ):
        G_simple = self.convert_multigraph_to_simple_graph(G=G, agg_weight_by=agg_weight_by)

        elarge = [(u, v) for (u, v, d) in G_simple.edges(data=True) if d["weight"] > weight_large_thr]
        esmall = [(u, v) for (u, v, d) in G_simple.edges(data=True) if d["weight"] <= weight_large_thr]

        pos = nx.spring_layout(G_simple, seed=7)  # positions for all nodes - seed for reproducibility
        # nodes
        nx.draw_networkx_nodes(G_simple, pos, node_size=draw_node_size)
        # edges
        nx.draw_networkx_edges(
            G_simple, pos, edgelist=elarge, width=draw_line_width
        )
        nx.draw_networkx_edges(
            G_simple, pos, edgelist=esmall, width=draw_line_width, alpha=0.5, edge_color="b", style="dashed"
        )
        # node labels
        nx.draw_networkx_labels(G_simple, pos, font_size=draw_font_size, font_family="sans-serif")
        # edge weight labels
        edge_labels = nx.get_edge_attributes(G_simple, "weight")
        nx.draw_networkx_edge_labels(G_simple, pos, edge_labels)

        ax = plt.gca()
        ax.margins(0.08)
        plt.axis("off")
        plt.tight_layout()
        plt.show()
        return


if __name__ == '__main__':
    Pandas.increase_display()
    lgr = Logging.get_default_logger(log_level=logging.DEBUG, propagate=False)
    gu = GraphUtils(logger=lgr)
    G = gu.create_multi_graph(
        edges = [
            {'key': 'plane', 'u': 'Shanghai', 'v': 'Tokyo', 'distance': 10, 'comment': 'Shanghai-Tokyo flight'},
            # duplicate (will not be added), order does not matter
            {'key': 'plane', 'u': 'Tokyo', 'v': 'Shanghai', 'distance': 22, 'comment': 'Tokyo-Shanghai flight'},
        ],
        col_weight = 'distance',
    )
    print(G)
    exit(0)
